/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "magic.h"
#include "gpc.h"
#include "gmdh.h"
#include "prob.h"
#include "proto.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"


#define xabs(x) ((x)>0?(x):(-(x)))

void evaluate_fitness_of_populations(
     int 	numpops,
     int 	numgens,
     pop_struct	*pop,
     int	p
     )
{
  int		i;

  for (i=0; i<pop[p].population_size; i++) {
    pop[p].standardized_fitness[i] =
      evaluate_fitness_of_individual(pop,p,pop[p].population[i],i);
  }
}


/***********:
* MSE claus - For great justice
*/

double eval_claus(tree *t, int i, double *values)
{
  double args[2], *ans, tmp;
  struct nonterminal_node *np;
  double *(*fn)(double *args);

  if (t->nodetype == TERMINAL)
  {
    if ((i - (1+t->id) < 0)||
        (i - (1+t->id) > i))
    {
      return 0;
    }
    else
    {
      if (values[i - (1+t->id)] > P.overlimit) return P.overlimit;
      if (values[i - (1+t->id)] < -1*P.overlimit) return -1*P.overlimit;
      return (values[i - (1+t->id)]);
    }
  }

  np = t->type.func->node_pointer;

  args[0]=eval_claus(t->type.func->arg[0],i,values);
  args[1]=eval_claus(t->type.func->arg[1],i,values);

  fn = function_code_from_type(np->type);
  ans = (*fn)(args);

  tmp = 0.0;

  for(int j=0; j<no_coeff_from_type(np->type);j++)
      tmp=tmp+(np->coeff[j])*ans[j];

  deallocvct(ans);

  return(tmp);
}

double mse_claus(tree *t, int terms)
{
   double *values;
   double sum;

   sum = 0.0;

   values = (double *)malloc(P.data_train_max*sizeof(double));
   for (int i=0; i<terms; i++)
     values[i] = P.data[i];
   for (int i=terms; i < P.data_train_max; i++)
     {
       values[i] = eval_claus(t,i,values);
       sum += (values[i]-P.data[i])*(values[i]-P.data[i]);
     }

   free(values);

   return (sum/(P.data_train_max - terms));

}



double evaluate_fitness_of_individual(
     pop_struct	*pop,
     int	p,
     tree	*t,
     int	i
     )
{
/* return mdl value */
  double mdlvalue,mse;

  // The second evaluation function is the correct one,
  // but the first does some initializing that I still haven't
  // found out what.
  mse = eval(t);
  //mse = mse_claus(t,pop[p].terminal_table_size);

  mdlvalue = mdl(t,mse,P.data_train_max);

  pop[p].mse[i]=mse;
  pop[p].mdl[i]=mdlvalue;

  return mdlvalue;
}

/* TODO: This function does nothing? */
int terminate_early(
  int 		numpops,
  int 		numgens,
  pop_struct 	*pop
  )

{
  int	p,i;

/*  for (p=0; p<numpops; p++) {
    for (i=0; i<pop[p].population_size; i++) {
      if (pop[p].standardized_fitness[i] <= 0.0) {
	return 1;
      }
    }
  } */
  return 0;
}
