/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/


/*
 *
 * $Log: generations.cpp,v $
 * Revision 1.2  2006/09/03 09:33:28  caranha
 * Addressing Identical Trees Problem - Not fully fixed yet
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.6  2006/01/26 15:55:13  caranha
 * code cleanup, output() bug fixed (i think)
 *
 * Revision 1.5  2006/01/26 10:09:29  caranha
 * added tree printing to Result
 *
 * Revision 1.4  2006/01/25 11:18:11  caranha
 * Added class for parameter manipulation. Moved some globals around.
 *
 * Revision 1.3  2006/01/25 05:20:51  caranha
 * Broke the generation loop - making parameter object
 *
 * Revision 1.2  2006/01/24 13:30:59  caranha
 * Modified "generations" to enable generation-per-generation running. Now you can stop one generation in order to pick it apart.
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.13  1993/04/22  07:39:12  gpc-avc
 * Removed old log messages
 *
 * Revision 2.12  1993/04/15  09:10:10  gpc-avc
 * Added bsd_qsort()
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "magic.h"
#include "gpc.h"
#include "prob.h"
#include "proto.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

#include "treeio.h"


/* report functions */
/* 1994 3 11 */

void initreport(
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st
  )
{
  char  ckptfile[256];
  FILE	*cf;

  /* Change by Claus 01/08 :
   * This will be later changed in the GUI, so I'm removing
   * the gethostid and getpid calls for the time
   */

  sprintf(ckptfile,"%s_%ld_%d.ckpt",st,42,42);  //,gethostid(),getpid());

  cf = fopen(ckptfile,"a");
/* file is created by startup for write_header */

  fprintf(cf, "random_seed = %ld\n",get_seed());
  fprintf(cf, "number_of_populations = %d\n",numpops);
  fprintf(cf, "number_of_generations = %d\n",numgens);
  fprintf(cf, "demes = %d\n",demes);
  fprintf(cf, "number_of_deme_rows = %d\n",demerows);
  fprintf(cf, "number_of_deme_columns = %d\n",demecols);
  fprintf(cf, "gaussian_noise_uniform1 = %22.20f\n",gaussian_noise_uniform1);
  fprintf(cf, "gaussian_noise_uniform2 = %22.20f\n",gaussian_noise_uniform2);
  fprintf(cf, "gaussian_noise_temp = %22.20f\n",gaussian_noise_temp);
  fprintf(cf, "gaussian_noise_toggle = %d\n",gaussian_noise_toggle);


  fclose(cf);
}


/* report utility */

/* (defun page (st) (format st "~%~|~%")) */

void page(FILE	*cf)
{
  fprintf(cf,"\n");
}

void repchar(FILE *outst,int  ch, int repcount)
{
  int j;
  for (j = 0; j < repcount; j++)
    putc(ch, outst);
}

VOID skip(FILE *outst, int  skipcount)
{
  int j;
  for (j = 0; j < skipcount; j++)
    fprintf(outst, "\n");
}

VOID write_nodes_of_tree (
  FILE	*cf,
  tree  *t
  )
{
  struct nonterminal_node *temp;
  struct no_tag *tagp;
  int i;

  if (t->nodetype == TERMINAL)
    return;
  else {
    temp = t -> type.func -> node_pointer;
    fprintf(cf, "NODE%d : type= %d ", temp, temp->type);
    if (temp->mse >= 0.0) fprintf(cf, "SN2 = %e ptr.= ", temp->mse);
/*  pop index not written */
/*
    tagp = temp -> pop_pointer;
    while (!(tagp == NULL)){
          fprintf(cf, "pop[%d][%d] ", tagp->popno, tagp-> index);
	  tagp = tagp -> next;
	}
    fprintf(cf, "\n");
*/
    fprintf(cf, "\n");
    fprintf(cf, "Coeff. ");
    for (i=0; i< no_coeff_from_type(temp->type) ; i++)
      fprintf(cf, "%e ", (temp->coeff)[i]);
    fprintf(cf, "\n");
    fprintf(cf, "Xlist. ");
    for (i=0; i<P.data_train_max; i++)
      fprintf(cf, "%3.10f ", (temp->xlist)[i]);
    fprintf(cf, "\n");

/* only root node written */
/*    write_nodes_of_tree(cf,t->type.func->arg[0]);
    write_nodes_of_tree(cf,t->type.func->arg[1]); */
  }
}




VOID write_nodes (FILE	*cf)
{
  struct nonterminal_node *temp;
  struct no_tag *tagp;
  int i;

  temp = node_list;

  while (!(temp == NULL)) {
    fprintf(cf, "NODE%d : type= %d ", temp, temp->type);
    if (temp->mse >= 0.0) fprintf(cf, "SN2 = %e ptr.= ", temp->mse);
    tagp = temp -> pop_pointer;
    while (!(tagp == NULL)){
          fprintf(cf, "pop[%d][%d] ", tagp->popno, tagp-> index);
	  tagp = tagp -> next;
	}
    fprintf(cf, "\n");
    fprintf(cf, "Coeff. ");
    for (i=0; i< no_coeff_from_type(temp->type) ; i++)
      fprintf(cf, "%e ", (temp->coeff)[i]);
    fprintf(cf, "\n");
    fprintf(cf, "Xlist. ");    
    for (i=0; i<P.data_train_max; i++)
      fprintf(cf, "%3.10f ", (temp->xlist)[i]);
    fprintf(cf, "\n");
    temp = temp -> next;
  }
}



VOID report_only_pop (
  int           pnum,
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st,
  int           flag
  )
{

  int	i, index;
  char  ckptfile[256];
  FILE	*cf, *co, *co2;
  float	sum = 0.0;

  /* Changed by Claus on 01/08 for porting to Borland c++*/
  sprintf(ckptfile,"%s_%ld_%d.ckpt",st,42,42);  //gethostid(),getpid());
    
  cf = fopen(ckptfile,"a");

  fprintf(cf, "<< Population No. %d at gen. %d >>\n", pnum, g);


  fprintf(cf, "gen %d : ncross_at_funct_pt= %d ncross_at_any_pt = %d nmut = %d\n",
	  g,
	  ncross_at_funct_pt,
	  ncross_at_any_pt,
	  nmut);
  
/* from dump_generation */
    if (flag){
      for (i=0; i<pop[pnum].population_size; i++) {
	fprintf(cf, "\n");    fprintf(cf, "\n");
	index = i;
	fprintf(cf, "pop no.= %d sn2 = %e mdl = %e standardized = %e, adjusted = %e, norm = %e\n",
		index,
		pop[pnum].mse[index],
		pop[pnum].mdl[index],
		pop[pnum].standardized_fitness[index],
		pop[pnum].adjusted_fitness[index],
		pop[pnum].normalized_fitness[index]);
	
	fprintf(cf, "pop no.= %d parent1 = %d parent2 = %d cross_funct = %d cross_any = %d mut = %d\n",
		index,
		pop[pnum].parent1[index],
		pop[pnum].parent2[index],
		pop[pnum].cross_funct[index],
		pop[pnum].cross_any[index],
		pop[pnum].mut_pos[index]);
	write_tree(pop,pop[pnum].population[index],pop[pnum].ckpt_format,cf);
      }
    }

/* from report_on_generation */
    sum = 0.0;

    for (i=0; i<pop[pnum].population_size; i++) {
      sum += pop[pnum].mse[i];
    }
    
    fprintf(cf, "\nGeneration %d Population %d   Avg Std SN2: %e\n",
	    g, pnum, sum/(float)pop[pnum].population_size);
    fprintf(cf,"Best-of-gen %d pop %d index %d sn2: %e mdl: %e\nBest-of-gen tree:\n",
	    g, pnum, pop[pnum].fitness_sort_index[0], 
	    pop[pnum].best_of_gen_mse,pop[pnum].best_of_gen_mdl);
    write_tree(pop,pop[pnum].best_of_generation,pop[pnum].ckpt_format,cf);
    //add2005/12/03
    co=fopen("resul_structure.txt","w");
    fprintf(co,"Best tree of Generation %d\n",g);
    fprintf(co,"MSE = %lf  MDL = %lf\n\n ",pop[pnum].best_of_run_mse,pop[pnum].best_of_run_mdl);
    write_tree_a(pop,pop[pnum].best_of_generation,pop[pnum].ckpt_format,co);
    fclose(co);
    //--------------------
    write_tree_a(pop,pop[pnum].best_of_generation,pop[pnum].ckpt_format,stdout);//add11/24
    printf("Best_MSE=%lf\n",pop[pnum].best_of_run_mse);//add11/24
    printf("Best_MDL=%lf\n",pop[pnum].best_of_run_mdl);//add11/24
    printf("Above Generation is %d\n-----------------------------------\n",g);//add11/24
    //add2005/1205
    if(g==0){
    co2 = fopen("result_generation.txt","w");
    }
    else{
    co2 = fopen("result_generation.txt","a");
    }
    fprintf(co2,"\nGeneration %d -------------------\n",g);
    write_tree_a(pop,pop[pnum].best_of_generation,pop[pnum].ckpt_format,co2);
    fprintf(co2,"\nBest_MSE=%lf\n",pop[pnum].best_of_run_mse);
    fprintf(co2,"Best_MDL=%lf\n\n",pop[pnum].best_of_run_mdl);
    fclose(co2); 

    
  if(flag) write_nodes(cf);
   else write_nodes_of_tree(cf,pop[pnum].best_of_generation);


  fclose(cf);
}

VOID report (
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st,
  int           flag
  )
{

  int	i, p, index;
  char  ckptfile[256];
  FILE	*cf;
  float	sum = 0.0;

  /* Removing Unix gethostid functions for Borland Compatibility */
  /* By Claus, 01/08 */
  sprintf(ckptfile,"%s_%ld_%d.ckpt",st,42,42); //gethostid(),getpid());

  cf = fopen(ckptfile,"a");


  for (p=0; p<numpops; p++) {

    fprintf(cf, "<< Population No. %d at gen. %d >>\n", p, g);


  fprintf(cf, "gen %d : ncross_at_funct_pt= %d ncross_at_any_pt = %d nmut = %d\n",
	  g,
	  ncross_at_funct_pt,
	  ncross_at_any_pt,
	  nmut);

/* from dump_generation */
    if (flag){
      for (i=0; i<pop[p].population_size; i++) {
	fprintf(cf, "\n");    fprintf(cf, "\n");
	/*     index = (DEMES? i : pop[p].fitness_sort_index[i]);*/
	index = i;
	fprintf(cf, "pop no.= %d sn2 = %e mdl = %e standardized = %e, adjusted = %e, norm = %e\n",
		index,
		pop[p].mse[index],
		pop[p].mdl[index],
		pop[p].standardized_fitness[index],
		pop[p].adjusted_fitness[index],
		pop[p].normalized_fitness[index]);

	fprintf(cf, "pop no.= %d parent1 = %d parent2 = %d cross_funct = %d cross_any = %d mut = %d\n",
		index,
		pop[p].parent1[index],
		pop[p].parent2[index],
		pop[p].cross_funct[index],
		pop[p].cross_any[index],
		pop[p].mut_pos[index]);
	write_tree(pop,pop[p].population[index],pop[p].ckpt_format,cf);
      }
    }

/* from report_on_generation */
    sum = 0.0;

    for (i=0; i<pop[p].population_size; i++) {
      sum += pop[p].mse[i];
    }

    fprintf(cf, "\nGeneration %d Population %d   Avg Std SN2: %e\n",
	    g, p, sum/(float)pop[p].population_size);
    fprintf(cf,"Best-of-gen %d pop %d index %d sn2: %e mdl: %e\nBest-of-gen tree:\n",
	    g, p, pop[p].fitness_sort_index[0],
	    pop[p].best_of_gen_mse,pop[p].best_of_gen_mdl);
    write_tree(pop,pop[p].best_of_generation,pop[p].ckpt_format,cf);

  if(flag) write_nodes(cf);
   else write_nodes_of_tree(cf,pop[p].best_of_generation);


  }


  fclose(cf);
}

/* gc functions */
VOID free_tag(struct no_tag *tagp)
{
  if (!(tagp == NULL)) {
    free_tag(tagp->next);
    free(tagp);
  }
}


VOID clear_pointer()
{
  struct nonterminal_node *temp;
  temp = node_list;

  while (!(temp == NULL)) {
    free_tag(temp->pop_pointer);
    temp -> pop_pointer = NULL;
    temp = temp -> next;
  }
}

VOID mark_pointer_by_tree(tree *tr, int p, int i, pop_struct *pop)
{
  struct no_tag *newtag;
  int j,k;
  
  if (tr->nodetype == FUNCTION) {
	for (j=0; j<function_arity(tr); j++) 
	  mark_pointer_by_tree(tr->type.func->arg[j],p,i,pop);
   if ((newtag =
	 (struct no_tag *) malloc(function_arity(tr) * sizeof(struct no_tag))) == NULL) {
      perror(" at point 1 in mark_pointer() <generations.c>");
      exit(errno);
    }
    newtag -> next = tr -> type.func -> node_pointer -> pop_pointer;
    tr -> type.func -> node_pointer -> pop_pointer = newtag;
    newtag -> popno = p;
    newtag -> index = i;	
  }
}

VOID mark_pointer(pop_struct *pop, int numpops)
{
  int i,p;
  for (p=0; p<numpops; p++) 
    for (i=0; i<pop[p].population_size; i++) 
      mark_pointer_by_tree(pop[p].population[i],p,i,pop);
}

VOID free_node(struct nonterminal_node *np)
{
  free(np->coeff);
  free(np->xlist);
  free(np);
}


VOID sweep_pointer()
{
  struct nonterminal_node *temp, *newtemp;

  temp = node_list;

  node_list = NULL;

  while (!(temp == NULL))
    if (!(temp->pop_pointer == NULL))
	{
	  newtemp = temp -> next;
	  temp -> next = node_list;
	  node_list = temp;
	  temp = newtemp;
	} else
	{
	  newtemp = temp -> next;
	  free_node(temp);
	  temp = newtemp;

	}
}

VOID gc_nodes(pop_struct *pop,int numpops)
{
  clear_pointer();
  mark_pointer(pop, numpops);
  sweep_pointer();
}  


VOID generations(
  int		numpops,
  int		numgens,
  int		start_gen,
  pop_struct	*pop,
  int		demes,
  pop_struct	***grid,
  int		demerows,
  int		demecols
  )
{
  int	g, p, i, j;
/*  int progress; */
  
  initreport(numpops, numgens, demes, demerows, demecols, pop, start_gen, "dat");
  initreport(numpops, numgens, demes, demerows, demecols, pop, start_gen, "cas");
  
  /* GENERATION */
  for (g=start_gen; g<numgens; g++)
    {
      /* progress = 0; */
      for (p = 0; p<numpops; p++)
	{
	  for (i=0; i<pop[p].population_size; i++)
	    {
	      pop[p].parent1[i] = -1;
	      pop[p].parent2[i] = -1;
	      pop[p].cross_funct[i] = -1;
	      pop[p].cross_any[i] = -1;
	      pop[p].mut_pos[i] = -1;
	    }
	}
      if (g) // not first generation
	{
	  for (p = 0; p<numpops; p++)
	    {
	      if (demes)
		{
		  for (i=0;i<demerows;i++)
		    {
		      for (j=0;j<demecols;j++)
			{
			  if (pop[p].steady_state)
			    {
			      /* not strictly necessary, really, but removes what
				 i think would otherwise be a trend for the upper
				 left hand of the grid to have much lower fitness
				 than the lower right */
			      breed_new_population(grid[random_int(demerows)]
						   [random_int(demecols)],
						   p,demes,demerows,demecols,
						   pop[p].steady_state);
			    }
			  else // !pop[p].steady_state
			    {
			      breed_new_population(grid[i][j],p,demes,demerows,demecols,
						   pop[p].steady_state);
			    }
			}
		    }
		}
	      else // !demes
		{
		  breed_new_population(pop,p,0,0,0,pop[p].steady_state);
		}
	      if (!(pop[p].steady_state))
		{
		  free_population(pop,p);
		  load_new_population(pop,p);
		}
	    }
	} else // !g -> first generation
	  {
	    /* generation 0 is random */
	    initialize_populations(numpops,pop);
	    for (p = 0; p<numpops; p++)
	      {
		if (pop[p].steady_state)
		  {
		    evaluate_fitness_of_populations(numpops,numgens,pop,p);
		    if (pop[p].parsimony_factor > 0.0) add_parsimony_to_fitness(pop,p);
		  }
	      }
	  } // finish running the population for this generation.
      
      /* NOTE that we breed ALL populations before we evaluate their fitness
	 (except in the case of steady-state).  This is done to support
	 co-evolution where there are multiple interacting populations
	 who are co-evaluated at each generation */
      
      for (p = 0; p<numpops; p++)
	{
	  if (!(pop[p].steady_state))
	    {
	      zero_fitness_of_populations(numpops,pop,p);
	      evaluate_fitness_of_populations(numpops,numgens,pop,p);
	      if (pop[p].parsimony_factor > 0.0) add_parsimony_to_fitness(pop,p);
	    }
	  
	  normalize_fitness_of_population(pop,p);
	  
	  /*      sort_population_by_fitness(pop,p); */
#if MDL_version
	  sort_population_by_mdl(pop,p);
#else
	  sort_population_by_mse(pop,p);
#endif
	  
	  
	  pop[p].best_of_gen_mse =
	    pop[p].mse[pop[p].fitness_sort_index[0]];
	  pop[p].best_of_gen_mdl =
	    pop[p].mdl[pop[p].fitness_sort_index[0]];
	  
	  pop[p].best_of_generation =
	    copy_tree(pop[p].population[pop[p].fitness_sort_index[0]]);
	  
	  
	  report_on_generation(g,pop,p);
	  
#if ((DEBUG == 1)||(DBSS == 1))
	  dump_population(pop,p);
#endif
#if ((DBDEMES == 1)||(DBSS == 1))
	  if (demes) { /* this should alway be true, but still.... */
	    for (i=0;i<demerows;i++) {
	      for (j=0;j<demecols;j++) {
		printf("\nDUMPING DEME row=%d col=%d\n",i,j);
		dump_population(grid[i][j],p);
	      }
	    }
	  }
#endif
	  
	  
	  if (!g)
	    {
	      pop[p].best_of_run_mdl = pop[p].best_of_gen_mdl ;
	pop[p].best_of_run_mse = pop[p].best_of_gen_mse ;
	pop[p].best_of_run = copy_tree(pop[p].population[pop[p].fitness_sort_index[0]]);
	pop[p].best_of_run_gen = 0;
	    }
	  //else if (((MDL_version == 1) && ( pop[p].best_of_gen_mdl < pop[p].best_of_run_mdl)) || ((MDL_version == 0) && ( pop[p].best_of_gen_mse < pop[p].best_of_run_mse)))
	  {
	    pop[p].best_of_run_mdl = pop[p].best_of_gen_mdl ;
	    pop[p].best_of_run_mse = pop[p].best_of_gen_mse ;
	    free((char *)pop[p].best_of_run);
	    pop[p].best_of_run = copy_tree(pop[p].best_of_generation);
	    pop[p].best_of_run_gen = g;
	    report_only_pop(p, numpops, numgens, demes, demerows, demecols, pop, g, "cas", DETAIL_CAS);
	  }//comment out 12/03  <- Shotgun Voodoo! Exactly what does this do, and why is it wrong?
	  
	}
      
      gc_nodes(pop,numpops);
      
      
      if ( CHECKPOINT_FREQUENCY ) {
	if ((!g) || !(g % CHECKPOINT_FREQUENCY)) {
	  report(numpops, numgens, demes, demerows, demecols, pop, g, "dat", DETAIL_DAT);
	}
      }
      
      if (g == start_gen)
	report(numpops, numgens, demes, demerows, demecols, pop, g, "cas", DETAIL_CAS);
      
      
      /* this is to be here for report */
      for (p = 0; p<numpops; p++)
	free_tree(pop[p].best_of_generation);
      
      if (terminate_early(numpops,numgens,pop)) break;
    } // end of one generation
  
  /* checkpoint the last generation, if it was not just saved */
  /*
    if (CHECKPOINT_FREQUENCY) {
    if ((numgens-1) % CHECKPOINT_FREQUENCY) {
    report(numpops,numgens,demes,demerows,demecols,pop,numgens-1,"dat",DETAIL_DAT);
    }
    } */
}


VOID dump_population(
  pop_struct 	*pop,
  int 		p
  )
{
  int	i, index;

  for (i=0; i<pop[p].population_size; i++) {
    index = (DEMES? i : pop[p].fitness_sort_index[i]);
    printf("pop= %d sn2 = %e mdl = %e standardized = %e, adjusted = %e, norm = %e\n",
	   p,
	   pop[p].mse[index],
	   pop[p].mdl[index],
	   pop[p].standardized_fitness[index],
	   pop[p].adjusted_fitness[index],
	   pop[p].normalized_fitness[index]);
    write_tree(pop,pop[p].population[index],pop[p].ckpt_format,stdout);
  }
}


VOID zero_fitness_of_populations(
  int		numpops,
  pop_struct	*pop,
  int		p
  )
{
  int	i;

  for (i=0; i<pop[p].population_size; i++) {
    pop[p].standardized_fitness[i] 	= 0.0;
    pop[p].adjusted_fitness[i] 		= 0.0;
    pop[p].normalized_fitness[i] 	= 0.0;
  }
}

int	global_p;

static int fitness_compare(const void *i, const void *j)
{
  pop_struct *pop = POP;

  if (pop[global_p].normalized_fitness[*((int *)j)] > pop[global_p].normalized_fitness[*((int *)i)]) {
    return 1;
  } else if (pop[global_p].normalized_fitness[*((int *)j)] < pop[global_p].normalized_fitness[*((int *)i)]) {
    return -1;
  } else return 0;
}

static int mse_compare(const void *i,const void *j)
{
  pop_struct *pop = POP;

/* the smaller, the better */

  if (pop[global_p].mse[*((int *)j)] < pop[global_p].mse[*((int *)i)]) {
    return 1;
  } else if (pop[global_p].mse[*((int *)j)] > pop[global_p].mse[*((int *)i)]) {
    return -1;
  } else return 0;
}


static int mdl_compare(const void *i, const void *j)
{
  pop_struct *pop = POP;

/* the smaller, the better */

  if (pop[global_p].mdl[*((int *)j)] < pop[global_p].mdl[*((int *)i)]) {
    return 1;
  } else if (pop[global_p].mdl[*((int *)j)] > pop[global_p].mdl[*((int *)i)]) {
    return -1;
  } else return 0;
}
    

VOID sort_population_by_fitness(
  pop_struct	*pop,
  int 	p
  )
{
  int	i;

  global_p = p; /* kludge for fitness_compare */
  for (i=0; i<pop[p].population_size; i++) {
    pop[p].fitness_sort_index[i] = i;
  }

  qsort(pop[p].fitness_sort_index, pop[p].population_size, sizeof(int),
	fitness_compare);

}

VOID sort_population_by_mse(
  pop_struct	*pop,
  int 	p
  )
{
  int	i;

  global_p = p; /* kludge for fitness_compare */
  for (i=0; i<pop[p].population_size; i++) {
    pop[p].fitness_sort_index[i] = i;
  }


  qsort(pop[p].fitness_sort_index, pop[p].population_size, sizeof(int),
	mse_compare);
}



VOID sort_population_by_mdl(
  pop_struct	*pop,
  int 	p
  )
{
  int	i;

  global_p = p; /* kludge for fitness_compare */
  for (i=0; i<pop[p].population_size; i++) {
    pop[p].fitness_sort_index[i] = i;
  }

  qsort(pop[p].fitness_sort_index, pop[p].population_size, sizeof(int),
	mdl_compare);

}


VOID add_parsimony_to_fitness(
  pop_struct 	*pop,
  int 		p
  )
{
  int	i;

  for (i=0; i<pop[p].population_size; i++) {
    pop[p].standardized_fitness[i] += 
      ((float) count_crossover_pts(pop[p].population[i]))*pop[p].parsimony_factor;
  }
}

VOID normalize_fitness_of_population(
  pop_struct	*pop,
  int 		p
)
{
  float	sum = 0.0;
  int	i;

  for (i=0; i<pop[p].population_size; i++) {
    sum += (pop[p].adjusted_fitness[i] = 1.0/(1.0 + pop[p].standardized_fitness[i]));
  }

  for (i=0; i<pop[p].population_size; i++) {
    pop[p].normalized_fitness[i] = (pop[p].adjusted_fitness[i]/sum);
  }
}



VOID report_on_generation(
  int 		g,
  pop_struct	*pop,
  int 		p
  )
{

  printf("\nGeneration %d Population %d\n", g, p);

}

