/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "gpc.h"
#include "malloc4.h"
#include "proto.h"

#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

  /* FIXME: THE FUNCTIONS TO WRITE TO FILE AND READ FROM FILE PARAMETERS
     DO NOT INCLUDE ALL THE CURRENT PARAMETERS */


// class constructor, sets some sane default values
parameters::parameters()
{
  numpops = 1; // number of populations
  numgens = 100; // number of generations

  eval_by_mdl = 1; //fitness: 1- eval by mdl, 0- eval by mse
  chkpt_freq = 1;  // every n gens write checkpoint file. 0 for none

  seed = 42; // random seed  - CANT BE ZERO! (why?)

   //

  weight = 1.0;
  weight_error = 1.0;

  for (int i = 0; i < POPMAX; i++)
  {
    terminal_table_size[i] = 100;
    population_size[i] 		= 100;
    steady_state[i]			= FALSE;
    strcpy(load_from_file[i]," ");
    max_depth_for_new_trees[i] 	= 6;
    max_depth_after_crossover[i] 	= 17;
    max_mutant_depth[i] 		= 10;
    grow_method[i] 			= GROW;
    selection_method[i] 		= TOURNAMENT;
    elite_size[i]                       = 2;
    tournament_K[i] 			= 5;
    deme_search_radius_sigma[i]	= 1.0;
    crossover_func_pt_fraction[i] 	= 0.4;
    crossover_any_pt_fraction[i] 	= 0.4;
    fitness_prop_repro_fraction[i]	= 0.0;
    parsimony_factor[i] 		= 0.0;
    func_gen_fraction[i] = 0.8;
    elite_dupe[i] = 0;
    crossover_dupe[i] = 1;
  }
}

// sets the global parameters of the object to their respective
// places.
void parameters::set_global_param()
{
  set_seed(seed);
  P.numpops = numpops;
  P.numgens = numgens;
  P.eval_by_mdl = eval_by_mdl;
  P.chkpt_freq = chkpt_freq;
  P.weight = weight;
  P.weight_error = weight_error;
}

// set parameters for all populations
void parameters::set_local_param()
{
  if (P.numpops == 0)
  {
    printf("No empty populations!\n");
    exit(0);
  }
  if (P.numpops >= POPMAX)
  {
    printf("Error: Number of populations is greater than compiled maximum!\n");
    exit(0);
  }


  POP = (pop_struct *) malloc(P.numpops*sizeof(pop_struct));
  for (int i=0; i < P.numpops; i++)
  {
    /* test sanity of parameters */
    if (elite_size[i] > population_size[i])
    {
      printf("Error: Elite size must be smaller than population size!\n");
      exit(0);
    }

    /* set parameters */
    POP[i].terminal_table_size          = terminal_table_size[i];
    POP[i].population_size 		= population_size[i];
    POP[i].steady_state			= steady_state[i];
    POP[i].load_from_file               = (char *)malloc(100*sizeof(char));
    POP[i].load_from_file[0]            = '\0';
    POP[i].max_depth_for_new_trees 	= max_depth_for_new_trees[i];
    POP[i].max_depth_after_crossover 	= max_depth_after_crossover[i];
    POP[i].max_mutant_depth 		= max_mutant_depth[i];
    POP[i].grow_method 			= grow_method[i];
    POP[i].selection_method 		= selection_method[i];
    POP[i].elite_size                   = elite_size[i];
    POP[i].tournament_K 		= tournament_K[i];
    POP[i].deme_search_radius_sigma	= 1.0;
    POP[i].crossover_func_pt_fraction 	= crossover_func_pt_fraction[i];
    POP[i].crossover_any_pt_fraction 	= crossover_any_pt_fraction[i];
    POP[i].fitness_prop_repro_fraction	= fitness_prop_repro_fraction[i];
    POP[i].parsimony_factor 		= parsimony_factor[i];
    POP[i].elite_dupe                   = elite_dupe[i]; 
    POP[i].crossover_dupe               = crossover_dupe[i];
  }
}


/*
VOID getparams(
  int		argc,
  char		**argv,
  int 		*numpops,
  int		*numgens,
  int		*demes,
  int		*demerows,
  int		*demecols,
  int 		*start_gen,
  FILE		**ckpt_file,
  pop_struct 	**pop,
  pop_struct    ****grid
  )
{
  int	i,j;
  //int   arg_offset = 0;
  //int   argp = 1;
  //FILE  *paramfile;
  //char ckptfile[255];
  //static char usage[] =
  //"%s: Usage: %s [-d rows cols] npops ngens paramfile | 'none'... [seed]\n%s: Usage: %s -r ckpt_filename\n";

    for(i=0;i<*numpops;i++) {
      defaultparams(*pop,i);

    }





}

*/

VOID defaultparams(
  pop_struct 	*pop,
  int		i
  )
{
  FILE	*paramfile;

  //set_seed((unsigned long)11287);
  //CHECKPOINT_FREQUENCY			= 0;
  //START_GEN				= 0;

  //pop[i].population_size 		= 100;
  //pop[i].steady_state			= FALSE;
  //pop[i].load_from_file 		= (char *) malloc(132*sizeof(char));
  //pop[i].load_from_file[0] 		= '\0';
  //pop[i].max_depth_for_new_trees 	= 6;
  //pop[i].max_depth_after_crossover 	= 17;
  //pop[i].max_mutant_depth 		= 4;
  //pop[i].grow_method 			= GROW;
  //pop[i].selection_method 		= FITNESSPROP;
  //pop[i].tournament_K 			= 7;
  //pop[i].deme_search_radius_sigma	= 1.0;
  //pop[i].crossover_func_pt_fraction 	= 0.695;
  //pop[i].crossover_any_pt_fraction 	= 0.195;
  //pop[i].fitness_prop_repro_fraction	= 0.095;
  //pop[i].parsimony_factor 		= 0.0;

  //if ((paramfile = fopen("default.in","r")) != (FILE *)NULL) {
  //  readparams(pop,i,paramfile);
  //  fclose(paramfile);
  //}
}

VOID readparams(
  pop_struct 	*pop,
  int		i,
  FILE 		*f
  )
{
  char	buf[132], name[132], val[132];
  unsigned long seed;
  FILE *tmp;

  while (fgets(buf,132,f) != (char *)NULL) {
    if (buf[0] == '#') continue;
    sscanf(buf,"%s = %s", name, val);
    if (!strcmp("seed",name)) {
      sscanf(val,"%ld",&seed);
      set_seed(seed);
    } else if (!strcmp("steady_state",name)) {
      sscanf(val,"%d",&(pop[i].steady_state));
    } else if (!strcmp("population_size",name)) {
      sscanf(val,"%d",&(pop[i].population_size));
    } else if (!strcmp("load_from_file",name)) {
      sscanf(val,"%s",  pop[i].load_from_file);
      if ((tmp=fopen(pop[i].load_from_file,"r")) == (FILE *)NULL) {
	fprintf(stderr,"Error: load_from_file not found: %s\n",
		pop[i].load_from_file);
	exit(1);
      }
      else {
	fclose(tmp);
      }
    } else if (!strcmp("checkpoint_frequency",name)) {
      sscanf(val,"%d",&(CHECKPOINT_FREQUENCY));
    } else if (!strcmp("max_depth_for_new_trees",name)) {
      sscanf(val,"%d",&(pop[i].max_depth_for_new_trees));
    } else if (!strcmp("max_depth_after_crossover",name)) {
      sscanf(val,"%d",&(pop[i].max_depth_after_crossover));
    } else if (!strcmp("max_mutant_depth",name)) {
      sscanf(val,"%d",&(pop[i].max_mutant_depth));
    } else if (!strcmp("grow_method",name)) {
      if (!strcmp(val,"FULL")) {
	pop[i].grow_method = FULL;
      } else if (!strcmp(val,"GROW")) {
	pop[i].grow_method = GROW;
      } else if (!strcmp(val,"RAMPED")) {
	pop[i].grow_method = RAMPED;
      } else {
	fprintf(stderr,"readparams(): bad grow_method val name %s\n",val);
      }
    } else if (!strcmp("selection_method",name)) {
      if (!strcmp(val,"TOURNAMENT")) {
	pop[i].selection_method = TOURNAMENT;
      } else if (!strcmp(val,"OVERSELECT")) {
	pop[i].selection_method = OVERSELECT;
      } else if (!strcmp(val,"FITNESSPROP")) {
	pop[i].selection_method = FITNESSPROP;
      } else {
	fprintf(stderr,"readparams(): bad selection_method val name %s\n",val);
      }
    } else if (!strcmp("tournament_K",name)) {
      sscanf(val,"%d",&(pop[i].tournament_K));
    } else if (!strcmp("deme_search_radius_sigma",name)) {
      sscanf(val,"%f",&(pop[i].deme_search_radius_sigma));
    } else if (!strcmp("crossover_func_pt_fraction",name)) {
      sscanf(val,"%f",&(pop[i].crossover_func_pt_fraction));
    } else if (!strcmp("crossover_any_pt_fraction",name)) {
      sscanf(val,"%f",&(pop[i].crossover_any_pt_fraction));
    } else if (!strcmp("fitness_prop_repro_fraction",name)) {
      sscanf(val,"%f",&(pop[i].fitness_prop_repro_fraction));
    } else if (!strcmp("parsimony_factor",name)) {
      sscanf(val,"%f",&(pop[i].parsimony_factor));
    } else if (!strcmp("_END_OF_PARAMETERS_POPULATION",name)) {
      int popnum;
      sscanf(val,"%d",&popnum);
      if (popnum != i) {
	fprintf(stderr,"Warning: readparams() possible population mismatch\n");
      }
      break;
    } else {
      fprintf(stderr,"readparams(): bad param name %s\n",name);
    }
  }
}

#ifdef ANSI_FUNC

VOID writeparams(
  pop_struct 	*pop,
  int		p,
  FILE 		*f
  )
#else

VOID writeparams(pop,p,f)
  pop_struct 	*pop;
  int		p;
  FILE		*f;
#endif
{
  fprintf(f, "seed = %ld\n",                    get_seed());
  fprintf(f,"checkpoint_frequency = %d\n",	CHECKPOINT_FREQUENCY);
  fprintf(f,"population_size = %d\n",		pop[p].population_size);
  fprintf(f,"steady_state = %d\n",		pop[p].steady_state);
  if (pop[p].load_from_file[0] != '\0') {
    fprintf(f,"load_from_file = %s\n", 		pop[p].load_from_file);
  }
  fprintf(f,"max_depth_for_new_trees = %d\n",	pop[p].max_depth_for_new_trees);
  fprintf(f,"max_depth_after_crossover = %d\n",	pop[p].max_depth_after_crossover);
  fprintf(f,"max_mutant_depth = %d\n",		pop[p].max_mutant_depth);
  fprintf(f,"grow_method = ");
  switch(pop[p].grow_method) {
  case FULL:
    fprintf(f,"FULL\n");
    break;
  case GROW:
    fprintf(f,"GROW\n");
    break;
  case RAMPED:
    fprintf(f,"RAMPED\n");
    break;
  }
  fprintf(f,"selection_method = ");
  switch(pop[p].selection_method) {
  case TOURNAMENT:
    fprintf(f,"TOURNAMENT\n");
    break;
  case OVERSELECT:
    fprintf(f,"OVERSELECT\n");
    break;
  case FITNESSPROP:
    fprintf(f,"FITNESSPROP\n");
  }
  fprintf(f,"tournament_K = %d\n",
	  pop[p].tournament_K);
  fprintf(f,"deme_search_radius_sigma = %f\n",
	  pop[p].deme_search_radius_sigma);
  fprintf(f,"crossover_func_pt_fraction = %f\n",
	  pop[p].crossover_func_pt_fraction);
  fprintf(f,"crossover_any_pt_fraction = %f\n",
	  pop[p].crossover_any_pt_fraction);
  fprintf(f,"fitness_prop_repro_fraction = %f\n",
	  pop[p].fitness_prop_repro_fraction);
  fprintf(f,"parsimony_factor = %f\n",
	  pop[p].parsimony_factor);
}


void checkpoint (
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st
  )
{
  int	i, p;
  char  ckptfile[256], cmd_buf[256];
  FILE	*cf;
  static int fn=0;

  sprintf(ckptfile,"%s_%ld_%d.ckpt%d",st,42,42,fn); // gethostid(),getpid(),fn); 01/08 by Claus
  
  fn = (fn + 1) % 10;
  
  cf = fopen(ckptfile,"w");

  fprintf(cf, "checkpoint_for_generation =  %d\n", g);
  fprintf(cf, "random_seed = %ld\n",get_seed());
  fprintf(cf, "number_of_populations = %d\n",numpops);
  fprintf(cf, "number_of_generations = %d\n",numgens);
  fprintf(cf, "demes = %d\n",demes);
  fprintf(cf, "number_of_deme_rows = %d\n",demerows);
  fprintf(cf, "number_of_deme_columns = %d\n",demecols);
  fprintf(cf, "gaussian_noise_uniform1 = %22.20f\n",gaussian_noise_uniform1);
  fprintf(cf, "gaussian_noise_uniform2 = %22.20f\n",gaussian_noise_uniform2);
  fprintf(cf, "gaussian_noise_temp = %22.20f\n",gaussian_noise_temp);
  fprintf(cf, "gaussian_noise_toggle = %d\n",gaussian_noise_toggle);

  fprintf(cf,"\n");

  for (p=0; p<numpops; p++) {
    writeparams(pop, p, cf);
    fprintf(cf,"_END_OF_PARAMETERS_POPULATION = %d\n",p);
  }

  for (p=0; p<numpops; p++) {
    fprintf(cf,"_START_OF_POPULATION_ = %d\n",p);
    for (i=0; i<pop[p].population_size; i++) {
      write_tree(pop, pop[p].population[i], pop[p].ckpt_format, cf);
    }
  }

  for (p=0; p<numpops; p++) {
    fprintf(cf, "best_of_run_fitness = %22.20f\n",pop[p].best_of_run_fitness);
    fprintf(cf, "best_of_run_gen = %d\n",pop[p].best_of_run_gen);
    write_tree(pop, pop[p].best_of_run, pop[p].ckpt_format, cf);
  }

  fclose(cf);

  // 01/08 by Claus - Unix command calling to compress checkpoint files...
  // need to implement this again in GUI (or not)

  /* sprintf(cmd_buf,"%s %s &",COMPRESS,ckptfile);
  system(cmd_buf);*/
}

void recover_params (
  FILE		*cf,
  int           *numpops,
  int           *numgens,
  int           *demes,
  int           *demerows,
  int           *demecols,
  pop_struct	**pop,
  pop_struct    ****grid,
  int		*g
  )
{
  int 	p;
  int	i, j;
  unsigned long 	seed;
  char	buf[132], name[132], val[132];

  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("checkpoint_for_generation",name)) { 
    sscanf(val, "%d", g);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("random_seed",name)) { 
    sscanf(val, "%ld",&seed);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("number_of_populations",name)) { 
    sscanf(val, "%d",numpops);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("number_of_generations",name)) { 
    sscanf(val, "%d",numgens);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("demes",name)) { 
    sscanf(val, "%d",demes);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("number_of_deme_rows",name)) { 
    sscanf(val, "%d",demerows);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("number_of_deme_columns",name)) { 
    sscanf(val, "%d",demecols);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("gaussian_noise_uniform1",name)) { 
    sscanf(val, "%f",&gaussian_noise_uniform1);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("gaussian_noise_uniform2",name)) { 
    sscanf(val, "%f",&gaussian_noise_uniform2);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("gaussian_noise_temp",name)) { 
    sscanf(val, "%f",&gaussian_noise_temp);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }
  fscanf(cf,"%s = %s", name, val);
  if (!strcmp("gaussian_noise_toggle",name)) { 
    sscanf(val, "%d",&gaussian_noise_toggle);
  }
  else {
    fprintf(stderr,"Error reading checkpoint file.\n");
  }

  fscanf(cf,"\n");
  
  *pop = (pop_struct *) malloc(*numpops*sizeof(pop_struct));
  if (*demes) {
    *grid = (pop_struct ***) malloc2(*demerows,*demecols*sizeof(pop_struct *));
    for (i=0;i<*demerows;i++) {
      for (j=0;j<*demecols;j++) {
	(*grid)[i][j] = (pop_struct *) malloc(*numpops*sizeof(pop_struct));
      }
    }
  }

  for (p = 0; p < *numpops; p++) {

    //(*pop)[p].load_from_file = (char *) malloc(132*sizeof(char)); // commented out because change of format in pop_struct
    readparams(*pop, p, cf);
    (*pop)[p].load_from_file[0] = '\0'; /* don't load from file upon recovery */
  }

  set_seed(seed);

}

