#ifndef _getparams_h_
#define _getparams_h_


// This maximum number of populations is used only for
// the parameters class. Actual number of populations is
// allocated after the parameter num_pops is set.
// If you believe you might need more populations, just set this
// higher.
#define POPMAX 10




// This class hold all the configurable parameters
// it uses this parameters for configuring population
// and problem
// it is the gui equivalent to the getparams function.

class parameters
{
public:
  /* MEMBERS */
  // Global variables

  int numpops; // number of populations
  int numgens; // number of generations

  int eval_by_mdl; //fitness: TRUE- eval by mdl, FALSE- eval by mse
  int chkpt_freq;  // every n gens write checkpoint file. 0 for none

  unsigned long seed; // random seed

  double weight;
  double weight_error;


  // Local variables (all of them are vectors for multiple populations)
  // some of these local variables are not really used, but are kept while I don't cut them off the code

  int terminal_table_size[POPMAX]; // number of terminals used
  int population_size[POPMAX]; //number of individuals
  int steady_state[POPMAX];    // unused
  char load_from_file[POPMAX][132]; // unused
  int max_depth_for_new_trees[POPMAX]; // ?
  int max_depth_after_crossover[POPMAX]; // ?
  int max_mutant_depth[POPMAX]; // ?
  int grow_method[POPMAX]; // See GPC.h line 55 (GROW or RAMPED)
  int selection_method[POPMAX]; // See GPC.h line 59 (Tournament, Overselection, Proportional)
  int elite_size[POPMAX]; // beyond the selection method, copies the best X individuals to the next generation.
  int tournament_K[POPMAX]; // Tournament size
  int deme_search_radius_sigma[POPMAX]; // unused

  // allow duplicate parameters
  int elite_dupe[POPMAX]; // allow duplicates when doing elite strategy
  int crossover_dupe[POPMAX]; // allow duplicates when doing crossover


  // Crossover/ Mutation probability parameters

  float crossover_func_pt_fraction[POPMAX]; // chance of xover on function node
  float crossover_any_pt_fraction[POPMAX]; // chance of xover on any node
  float fitness_prop_repro_fraction[POPMAX]; // chance of copying the node intact

  // beyond the three probabilities above is the mutation chance.
  // The sum of the above three parameters should not be above 1.0 (nothing BAD will happen if it is, though)



  int parsimony_factor[POPMAX]; // This is added to fitness if > 0 for fitness calculation. Don't know the use.

  double func_gen_fraction[POPMAX]; // used when node is randomly generated. the fraction of function generation.

  /* METHODS */

  parameters(); //constructor sets default runnable values

  void set_global_param(); // Put the global parameters to their respective places
  void set_local_param(); // Put the local parameters in their respective POP vectors
};

#endif
