/* gmdh header  */
/* 1994 3 14 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "gpc.h"
#include "prob.h"
#include "clindef.h"
#include "gmdh.h"
#include "proto.h"

#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"




/* Globals */

/**
 * \fn double power(double x,int n)
 * \brief calculate x^n
 *
 * \param x base
 * \param n exponent
 * \return x^n
 */
double power(double x,int n)
{
  int i;
  double p;
  
  p=1.0;
  for (i=1; i<=n; ++i)
    p = p*x;
  return(p);
}




/**
 * \fn double mdl (tree *tr, double sn2, int n)
 * \brief calculate Minimum Description Length from the given tree
 *
 * \param tr tree as a genotype
 * \param sn2 mean square error
 * \param n number of data
 * \return MDL
 */
double mdl (tree *tr, double sn2, int n)
{
  int k;

  k=k_hat_new(tr);


  // Changing MDL definition to fit Nikolay & IBA 2001
  // OLD definition
  // return ( WEIGHT_ERROR*( n * log(sn2)) + WEIGHT * ( k * log(n)));
  
  // NEW definition
  
  return (WEIGHT_ERROR*sn2) + WEIGHT*((k*P.data_train_variance*log(n))/n);


}

int k_hat_new(tree *tr)
{
  struct nonterminal_node *np;

  if (tr->nodetype == TERMINAL) return(0);
  
  np = tr -> type.func -> node_pointer;
 
  if (np->type > 0)
    return( k_hat_new(tr->type.func->arg[0])
	   + k_hat_new(tr->type.func->arg[1])
	   + no_coeff_from_type(np->type));
    else return( k_hat_new(tr->type.func->arg[0])
		 + no_coeff_from_type(np->type));
}


double mean_sqr_error (tree *tp)
{
/* coeff, xlist (of tp -> np) have been set */

  int i;
  double sum = 0.0, x,f;
  double V; // Derivative value for subpolinomial
  struct nonterminal_node *np;

  /* VALOR AQUI DEVERIA SER de START INDEX ATE o TRAIN VAL */
  if (tp->nodetype == TERMINAL) {
    for (i=0; i<P.data_train_max; i++) {
      f = P.data[i];
      x = P.getTerminal(i, (1 + tp->id)); /* should be 1+ */
      sum = sum + (f-x)*(f-x);
    }
    return (sum/P.data_train_max);
  }
  
  np = tp -> type.func -> node_pointer;
  
  if (np->mse >=0.0) return(np->mse);

  // Value of partial derivative based on nodetype (16 types of 
  // polinomials (by Claus))
  // Table from Adaptive Learning book
  // partial derivative from Nicolay & IBA 2001
  switch(np->type)
    {
    case 1:
      V = power(2*np->coeff[3],2);
      break;
    case 2:
      V = 0;
      break;
    case 3: 
      V = power(2*np->coeff[2],2);
      break;
    case 4:
      V = power(2*np->coeff[2] + 2*np->coeff[3],2);
      break;
    case 5:
      V = power(2*np->coeff[2],2);
      break;
    case 6:
      V = power(2*np->coeff[3],2);
      break;
    case 7:
      V = power(2*np->coeff[2] + 2*np->coeff[3],2);
      break;
    case 8:
      V = power(2*np->coeff[1] + 2*np->coeff[2],2);
      break;
    case 9:
      V = power(2*np->coeff[3] + 2*np->coeff[4] + 2*np->coeff[5],2);
      break;
    case 10:
      V = power(2*np->coeff[3] + 2*np->coeff[4],2);
      break;
    case 11:
      V = power(2*np->coeff[2] + 2*np->coeff[3] + 2*np->coeff[4],2);
      break;
    case 12:
      V = power(2*np->coeff[1] + 2*np->coeff[2] + 2*np->coeff[3],2);
      break;
    case 13:
      V = power(2*np->coeff[2] + 2*np->coeff[3],2);
      break;
    case 14:
      V = power(2*np->coeff[3] + 2+np->coeff[4],2);
      break;
    case 15:
      V = power(2*np->coeff[1],2);
      break;
    case 16:
      V = power(2*np->coeff[1] + 2*np->coeff[2],2);
      break;
    }

  /* Fix initial value for this loop, to avoid zeroes */
  for (i=0; i < P.data_train_max; i++) {
    f = P.data[i];
    x = np -> xlist [i];
    sum = sum + (f-x)*(f-x);
  }

  // Normal Average error:
  // return (sum/P.data_train_max)


  // Regularized Average Error - RAE, as per Nikolay and IBA 2001
  // edited by Claus in 06/11/21
  return ((sum + P.reg_param*V)/P.data_train_max);
}

/**
 * \fn VOID xmake_by_coeff (tree *tp)
 *
 * \n param tp tree as a genotype
 */
VOID xmake_by_coeff (tree *tp)
{
  int i,j;
  double tmp;
  struct nonterminal_node *np;

  np = tp->type.func->node_pointer;
  
  if (!(np -> xlist == NULL)) deallocvct(np->xlist);
  np -> xlist = allocdvct(P.data_train_max);

  for (i=0; i<P.data_train_max; i++) {
    tmp = 0.0;
    for (j=0; j< no_coeff_from_type(np->type) ; j++)
      tmp = tmp + (np->coeff[j])*FAST_X_ARRAY[i][j];
    np->xlist[i] = tmp;
  }
}

/**
 * \fn VOID fast_x_array_make (double x1, double x2, int i, int p, double *(*fn)(GENERIC *args))
 * \brief calculate X array required by calculation of GMDH coefficients
 *
 * \param x1 x1 in the text book.(Adaptive Learning of Polynomial Networks pp.11, eqn(1.8))
 * \param x2 x2 in the text book.
 * \param i index of row of FAST_X_ARRAY
 * \param p number of columns of FAST_X_ARRAY
 * \param *(*fn)(GENERIC *args) function to calculate the values of FAST_X_ARRAY
 * \return void
 */
VOID fast_x_array_make (double x1, double x2, int i, int p, double *(*fn)(GENERIC *args))
{
  int j;
  double *ans, args[2];

  args[0] = x1;
  args[1] = x2;

  ans=(*fn)(args);
  for (j=0; j<p; j++)
    FAST_X_ARRAY[i][j] = ans[j];
  deallocvct(ans);
}


/**
 * \fn double *b_calc(int n, int p)
 * \brief calculate coefficients of GMDH
 *
 * \param n number of rows of FAST_X_ARRAY
 * \param p number of columns of FAST_X_ARRAY
 * \return coefficients of GMDH
 */
double *b_calc(int n, int p)
{
/* FAST_X_ARRAY has been set */
  int i, j;
  double tmp;
  
  double **x2, **x2x, *x2y, **inv, *ans;

  x2 = dmattran(FAST_X_ARRAY,n,p); ///< X2 = tran(X)

  x2x = dmatmlt(x2, p, n, FAST_X_ARRAY, n, p); ///< X2X = tran(X) * X
  
  x2y = allocdvct(p); ///< alloc double vector x2y

/* x2 y setting */
/// x2y = X2*y = tran(X)*y
  for (i=0; i< p; i++) {
    tmp = 0.0;
    for (j=0; j< n; j++) 
      tmp = tmp + x2[i][j] * P.getTerminal(j,0);
    x2y[i] = tmp;
  }

  /* Added by claus on 06/11/21
   * Sum k*I to x2x, where k is regularization parameter.
   * should have no effect if k is zero
   * This is according to Nikolay and Iba, 2001.
   */
  for (i = 0; i < p; i++)
    x2x[i][i] += P.reg_param;


  inv=pseudo(x2x, p, p); ///< inv = inv( tran(X)*X ) = pseudo inverse
  ans=dmatdvct(inv, p,p, x2y, p); /// < ans = inv( tran(X)*X ) * tran(X) * y
  
  deallocmat(x2x);
  deallocmat(inv);
  deallocmat(x2);
  deallocvct(x2y);

  return (ans);

}

#define REDUCE_THRESH 0.001
/* TODO: WARNING */
int reduce_judge(double *xl1, double *xl2, int n)
{
  int i;
  for (i=0; i<n; i++)
    if (abs(xl1[i]-xl2[i]) > REDUCE_THRESH)
      return (0);
  return (1);
}

#define rnd(i)  random_int(i)+1 /* return from 1 to i */

void fast_fit_gmdh(tree *t)
{
  int i,j,p;
  int min_index, tmp_index;
  double min_mse, tmp_mse;
  tree *left, *right;
  double x1,x2;
  double *xlist1,*xlist2;
  struct nonterminal_node *np;
  int reduce_flag;

  static int saitama=0;

  mytree=t;
  
  if (t->nodetype == TERMINAL) return;
  
  np = t->type.func->node_pointer;


  if (!(np->type == NO_TYPE) && !(np->xlist ==NULL) && !(np->coeff ==NULL)){
    return;
  }

  xlist1 = (double *)malloc(P.data_train_max*sizeof(double));
  xlist2 = (double *)malloc(P.data_train_max*sizeof(double));
  
  
  left = t->type.func->arg[0];
  right = (t->type.func->arg)[1];
  
  if ((left->nodetype == FUNCTION) &&
      (left->type.func->node_pointer->xlist)==NULL)
    fast_fit_gmdh (left);
  
  if ((right->nodetype == FUNCTION) &&
      (right->type.func->node_pointer->xlist)==NULL)
    fast_fit_gmdh(right);
  
  for (i=0; i<P.data_train_max; i++) {
    if (left->nodetype == FUNCTION)
      xlist1[i] = (left->type.func->node_pointer->xlist)[i];
    else xlist1[i] = P.getTerminal(i, (1 + left->id));  /* should be +1  */

    if (right->nodetype == FUNCTION)
      xlist2[i] = (right->type.func->node_pointer->xlist)[i];
    else xlist2[i] = P.getTerminal(i, (1+ right->id));  /* should be +1  */
  }

  reduce_flag = reduce_judge(xlist1,xlist2,P.data_train_max);
  bool flag = false;
  if ((np->type == NO_TYPE) && RANDOM_GMDH) {

    flag = true;

    np -> type = rnd(no_node_functions);
    if (reduce_flag) np -> type = - (np -> type);
  }

  if (!(np->type == NO_TYPE)) /* type is already set by user, or RANDOM_GMDH */
    {
      
      for (i=0; i<P.data_train_max; i++)
	fast_x_array_make (xlist1[i], xlist2[i], i, no_coeff_from_type(np->type), (*(function_code_from_tree(t))));
      if (!(np->coeff) == NULL)   deallocvct(np->coeff);
      np -> coeff = b_calc(P.data_train_max, no_coeff_from_type(np->type));
      xmake_by_coeff(t);
      np -> mse = mean_sqr_error(t);
      free(xlist1);
      free(xlist2);
      return;
    }
  


  free(xlist1);
  free(xlist2);
}

/* matrix from kurita san */

/************************************************************/
/*	print out error message				    */
/************************************************************/
void	error_message(char *s1, char *s2)
{
	char errbuf[256];
	printf("ERROR %s : %s\n",s1,s2);
}


/****************************************************************/
/*      allocate resion for a vector    			*/
/*==============================================================*/
/*	Usage: V = (double *)allocvct(m,siseof(double));	*/
/****************************************************************/
char	*allocvct(int m, int elmsize)
{
	long	alloc_size;
	char	*V;

	alloc_size = m;
	alloc_size = alloc_size*elmsize;	
	V = (char *)malloc(alloc_size);
	return(V);
}

/****************************************************************/
/*      allocate resion for a matrix 		     		*/
/*==============================================================*/
/*	Usage: A = (double **)allocmat(m,n,sizeof(double));	*/
/****************************************************************/
char	**allocmat(int m, int n, int elmsize)
{
	int	i;
	long	alloc_size;
	char    *ptr;
	char	**A;

	alloc_size = m+1;
	alloc_size = alloc_size*sizeof(char *);

	A = (char **)malloc(alloc_size);

	for (i=0; i < m; i++) {
		ptr = allocvct(n,elmsize);
		A[i] = ptr;
	}
	A[m] = NULL;

	return(A);
}

/************************************************************/
/*	allocate nbytes reasion on heap			    */
/************************************************************/
char *alloc(long nbytes)
{
char	*cpt;

	if ((cpt = (char *)malloc(nbytes)) == NULL) {
		printf("alloc: No more memory !!\n");
		exit(0);
		cpt = NULL;
	}
	return(cpt);
}



/* Dealloc functions altered by Claus 01/08 to use template types */
/************************************************************/
/*	release reasion					    */
/************************************************************/

template <class ttype>
void dealloc(ttype *p)
{
	if (p != NULL) {
		free(p);
	}
}


/********************************************************/
/*      deallocate resion for a vector V  		*/
/********************************************************/
template <class ttype>
void deallocvct(ttype *V)
{
	dealloc(V);
}


/********************************************************/
/*      deallocate resion for a matrix A   		*/
/********************************************************/
template <class ttype>
void	deallocmat(ttype **A)
{
	int	i,flag = TRUE;
	ttype	*ptr;
	
	if (A != NULL) {
		for(i=0; flag == TRUE; i++) {
			ptr = (ttype *)A[i];
			if (ptr == NULL) flag = FALSE;
			else {
				deallocvct(ptr);
			}
		}
		dealloc((ttype *)A);
	}
}

/**
 * \fn double **pseudo(double **A,int M, int N)
 * \brief calculate Pseudo Inverse Matrix
 *
 * \param A matrix
 * \param M number of rows
 * \param N number of columns
 * \return Pseudo Inverse of A
 **/
/***********************************************************
	Moore-Penrose	Pseudo Inverse	
	   <= A+   : n*m MATRIX	
============================================================
	Usage:	P = pseudo(A,M,N);
************************************************************
A:	Original Matrix A(M,N)
***********************************************************/
double **pseudo(double **A,int M, int N)
{
  double **B, **AA, **U, **V, **L, **T;
  double *EVL;
  int    i, j, k;
  double evl1, check;

  if (N <= M) {
    /* AA <= A'A (N*N) */
    AA = dmatmlt2(A,M,M,A,M,M);
    EVL = allocdvct(N); V = allocdmat(N,N);
    egntj(AA,N,EVL,V);
    deallocmat((char **)AA);	/* reuse of variable AA */
    L = dmatzer(N,N);
    evl1 = EVL[0];
    for (i = 0; i < N; i++) {
      check = EVL[i] / evl1;
      if (check >= REPS) L[i][i] = 1.0 / EVL[i];
      else break;
    }
/* #ifdef DEBUG
    printf("rank = %d\n",i);
#endif  */
    deallocvct((char *)EVL);
    /* T <= AV  (M*N) */
    T = dmatmlt(A,M,N,V,N,N);
    /* AA <= VL  (N*N) */
    AA = dmatmlt(V,N,N,L,N,N);
    deallocmat((char **)V); deallocmat((char **)L);
    /* B <= AAT' = VLT' (N*M) */
    B = dmatmlt3(AA,N,N,T,M,N);
    deallocmat((char **)AA); 
    deallocmat((char **)T); 
    return(B);
  } else {
    /* AA <= AA' (M*M) */
    AA = dmatmlt3(A,M,N,A,M,N);
    EVL = allocdvct(M); U = allocdmat(M,M);
    egntj(AA,M,EVL,U);
    deallocmat((char **)AA);
    L = dmatzer(M,M);
    evl1 = EVL[0];
    for (i = 0; i < M; i++) {
      check = EVL[i] / evl1;
      if (check >= REPS) L[i][i] = 1.0 / EVL[i];
      else break;
    }
#ifdef DEBUG
    printf("rank = %d\n",i);
#endif
    deallocvct((char *)EVL);
    /* T <= A'U (N*M) */
    T = dmatmlt2(A,M,N,U,M,M);
    /* AA <= LU' (M*M) */
    AA = dmatmlt3(L,M,M,U,M,M);
    deallocmat((char **)U); deallocmat((char **)L); 
    /* B <= TAA = TLU' (N*M) */ 
    B = dmatmlt(T,N,M,AA,M,M);
    deallocmat((char **)AA); deallocmat((char **)T);
    return(B);
  }
}


/************************************************/
/*	check pseudo inverse			*/
/*	X(M,N) : Original Matrix		*/
/*	GX(N,M) : Pseudo Inverse of X		*/
/************************************************/
void	check_ginv(double **X, int M, int N, double **GX)
{
  int    i,j;
  double **U, **V;
	
  /* check G-inverse */
  U = dmatmlt(X,M,N,GX,N,M); V = dmatmlt(U,M,M,X,M,N);
  for (i = 0; i < M; i++) {
    j = i % N;
    printf("i=%4d j=%4d : X = %15.5f  --> X(X+)X = %15.5f\n",
			i,j,X[i][j],V[i][j]);
  }
  deallocmat((char **)V); deallocmat((char **)U);

}


/********************************************************
 *	eigen problem AV=VD
 *		Threshold Jacobi Method			
 *======================================================
 *	Usage:	egntj(A,M,D,V);			
 * A(M,M) : Input Matrix
 * V(M,M) : Eigen Vectors
 * D(M)   : Eigen Values
 ********************************************************/
void	egntj(double **A, int M, double *D, double **V)
{
  double **AA, *B, *Z;
  int	 i, j, jx, ip, iq;
  double sm, tresh, g, h;
  double t, theta, c, s, tau;
  double dmax;
  double sqval;
  double w, EPS = 1.0E-10;

  AA = dmatcp(A,M,M);
  B = allocdvct(M); Z = allocdvct(M);

  /* set V = I */
  for (i = 0; i < M; i++) {
    for (j = 0; j < M; j++) {
      if (i == j) V[i][j] = 1.0; else V[i][j] = 0.0;
    }
  }

  /* set diag(AA) to D and B */
  for (ip = 0; ip < M; ip++) {
    D[ip] = B[ip] = AA[ip][ip]; Z[ip] = 0.0;
  }

  for (i = 1; i <= 10000; i++) {	/* loop1 */
    sm = 0.0;
    for (ip = 0; ip < M-1; ip++) {
      for (iq = ip+1; iq < M; iq++) sm += abs(AA[ip][iq]);
    }
    if (sm < EPS) break;
    if (i < 4) tresh = 0.2 * (sm / (M * M)); else tresh = 0.0;

    for (ip = 0; ip < M-1; ip++) {	/* loop2 */
      for (iq = ip+1; iq < M; iq++) {	/* loop3 */
        g = 100.0 * abs(AA[ip][iq]);
        if ((i > 4) && ((abs(D[ip]) + g) == abs(D[ip])) &&
                     ((abs(D[iq]) + g) == abs(D[iq]))) {
          AA[ip][iq] = 0.0;
        } else {
          if (abs(AA[ip][iq]) > tresh) {
            h = D[iq] - D[ip];
            if (abs(h) + g == abs(h)) {
              t = AA[ip][iq] / h;
            } else {
              theta = 0.5 * h / AA[ip][iq];
              sqval = 1.0 + theta * theta;
              t = 1.0 / (abs(theta) + sqrt(sqval));
              if (theta < 0.0) t = -t;
            }
            sqval = 1.0 + t * t;
            c = 1.0 / sqrt(sqval);
            s = t * c;
            tau = s / (1.0 + c);
            h = t * AA[ip][iq];
            Z[ip] -= h; Z[iq] += h; D[ip] -= h; D[iq] += h;
            AA[ip][iq] = 0.0;
            if (ip > 0) {
              for (j = 0; j < ip; j++) {
                g = AA[j][ip]; h = AA[j][iq]; 
                AA[j][ip] = g - s * (h + g * tau);
                AA[j][iq] = h + s * (g - h * tau);
              }
            }
            if (ip+1 <= iq-1) {
              for (j = ip+1; j <= iq-1; j++) {
                g = AA[ip][j]; h = AA[j][iq];
                AA[ip][j] = g - s * (h + g * tau);
                AA[j][iq] = h + s * (g - h * tau);
              }
            }
            if (iq+1 < M) {
              for (j = iq+1; j < M; j++) {
                g = AA[ip][j]; h = AA[iq][j];
                AA[ip][j] = g - s * (h + g * tau);
                AA[iq][j] = h + s * (g - h * tau);
              }
            }
            for (j = 0; j < M; j++) {
              g = V[j][ip]; h = V[j][iq];
              V[j][ip] = g - s * (h + g * tau);
              V[j][iq] = h + s * (g - h * tau);
            }
          }
        }
      }	/* end of loop3	*/
    } /* end of loop2 */
  } /* end of loop1 */

  for (i = 0; i < M; i++) {
    jx = i; dmax = -1.0e30;
    for (j = i; j < M; j++) {
      if (dmax < D[j]) {
        jx = j; dmax = D[j];
      }
    }
    if (jx != i) {
      dswap(&(D[i]),&(D[jx]));
      for (j = 0; j < M; j++) {
        dswap(&(V[j][i]),&(V[j][jx]));
      }
    }
  }
  deallocvct((char *)B); deallocvct((char *)Z); deallocmat((char **)AA);
}


#define EPS 1.0e-5

/********************************************************/
/*	generalized eigen problem Ax = uBx		*/
/*		use	eigenn				*/
/*======================================================*/
/*	Usage:	gegntj(A,B,m,EL,EV);			*/
/********************************************************/
void	gegntj(
     double **A,
     double **B,
     int m,
     double *EVALUE,
     double **EVECTOR)
{
  double **R, **U, **V, **S;
  double *EV;
  int    i,j;
  double re;

  EV = allocdvct(m); U = allocdmat(m,m);

  egntj(B,m,EV,U);

  for (i = 0; i < m; i++) {
    re = abs(EV[i]);
    if (re >= EPS) EV[i] = 1.0 / sqrt(re); else EV[i] = 0.0;
  }

  /* U <= U H^(-1/2) */
  for (i = 0; i < m; i++) {
    for (j = 0; j < m; j++) U[i][j] = U[i][j] * EV[j];
  }
  deallocvct((char *)EV);

  /* V <= U' */
  V = dmattran(U,m,m);

  /* R <= VAU */
  S = dmatmlt(A,m,m,U,m,m);
  R = dmatmlt(V,m,m,S,m,m);
  deallocmat((char **)V);

  egntj(R,m,EVALUE,S);

  V = dmatmlt(U,m,m,S,m,m);
  deallocmat((char **)S); deallocmat((char **)U); deallocmat((char **)R);
  for (i = 0; i < m; i++) {
    for (j = 0; j < m; j++) EVECTOR[i][j] = V[i][j];
  }
  deallocmat((char **)V);
}





/************************************************/
/*      B <= A'  (double matrix)       		*/
/*==============================================*/
/*	Usage: B = dmattran(A,m,n);		*/
/************************************************/
double  **dmattran(
		   double **A,
		   int m,
		   int n)
{
	double 	**B;
	int	i,j;

	B = allocdmat(n,m);
	if (B != NULL) {
		for (i = 0; i < n; i++) {
			for (j = 0; j < m; j++) B[i][j] = A[j][i]; 
		}
	}
	return(B);

}


/********************************************************/
/*      C <= A * B   (double matrix)     			*/
/*======================================================*/
/*	Usage: C = dmatmlt(A,ma,na,B,mb,nb);		*/
/********************************************************/
double **dmatmlt(
		 double **A,
		 int ma, int na,
		 double **B,
		 int mb, int nb)
{
  double **C;
  int	i,j,k;

  if (na == mb) {
    C = allocdmat(ma,nb);
    if (C != NULL) {
      for (i = 0; i < ma; i++) {
	for (j = 0; j < nb; j++) {
	  C[i][j] = 0.0;
	  for (k = 0; k < na; k++) {
	    C[i][j] += A[i][k] * B[k][j];
	  }
	} 
      }
    }
  } 
  else {
    error_message("rmatmlt","dimension inconsistent");
    printf("  A[%d][%d]  :  B[%d][%d] \n",ma,na,mb,nb);
    C = NULL;
  }
  return(C);
}


/***********************************************/
/*      double matrix processing              */
/***********************************************/
/****************************************/
/*       <= 0  (double matrix)		*/
/*======================================*/
/*	Usage: A = dmatzer(m,n);	*/
/****************************************/
double	**dmatzer(
		  int m, int n)
{
	double	**A;
	int	i,j;

	A = allocdmat(m,n);
	if (A != NULL) {
		for (i = 0; i < m; i++) {
			for (j = 0; j < n; j++) A[i][j] = 0.0; 
		}
	}
	return(A);
}

/**
 * \fn void dswap( double *a, double *b)
 * \breif exchange two double data
 *
 * \param *a double value 1
 * \param *b double value 2
 */
/********************************************************/
/*	exchange two double data			*/
/*======================================================*/
/*	Usage: dswap(&a,&b);				*/
/********************************************************/
void	dswap(
	      double *a, double *b)
{
  double w;

  w = *a; *a = *b; *b = w;
}


/**
 * \fn double	**dmatcp( double **A, int m, int n)
 * \breif copy Matrix
 *
 * \param *A original matrix
 * \param m number of rows
 * \param n number of columns
 * \return copied Matrix of A
 */
/************************************************/
/*      B <= A  (double matrix) 			*/
/*==============================================*/
/*	Usage: B = dmatcp(A,m,n);		*/
/************************************************/
double	**dmatcp(
		 double **A,
		 int m, int n)
{
	double	**B;
	int	i,j;

	B = allocdmat(m,n);
	if (B != NULL) {
		for (i = 0; i < m; i++) {
			for (j = 0; j < n; j++) B[i][j] = A[i][j]; 
		}
	} 
	return(B);
}



/********************************************************/
/*      C <= A' * B   (double matrix)     			*/
/*======================================================*/
/*	Usage: C = dmatmlt2(A,ma,na,B,mb,nb);		*/
/********************************************************/
double **dmatmlt2(
		  double **A,
		  int ma, int na,
		  double **B,
		  int mb, int nb)
{
  double **C;
  int	i,j,k;

  if (ma == mb) {
    C = allocdmat(na,nb);
    if (C != NULL) {
      for (i = 0; i < na; i++) {
	for (j = 0; j < nb; j++) {
	  C[i][j] = 0.0;
	  for (k = 0; k < ma; k++) {
	    C[i][j] += A[k][i] * B[k][j];
	  }
	} 
      }
    }
  } 
  else {
    error_message("rmatmlt2","dimension inconsistent");
    printf("  A[%d][%d]  :  B[%d][%d] \n",ma,na,mb,nb);
    C = NULL;
  }
  return(C);
}


/**
 * \fn double **dmatmlt3( double **A, int ma, int na, double **B, int mb, int nb)
 * \brief multiply Matrix and Matrix (A*B)
 * \param A matrix
 * \param ma number of rows of A
 * \param na number of columns of A
 * \param B matrix
 * \param mb number of rows of B
 * \param nb number of columns of B
 * \return the matrix A*B
 */

/********************************************************/
/*      C <= A * B'   (double matrix)     		*/
/*======================================================*/
/*	Usage: C = dmatmlt3(A,ma,na,B,mb,nb);		*/
/********************************************************/
double **dmatmlt3(
		  double **A,
		  int ma, int na,
		  double **B,
		  int mb, int nb)
{
  double **C;
  int	i,j,k;

  if (na == nb) {
    C = allocdmat(ma,mb);
    if (C != NULL) {
      for (i = 0; i < ma; i++) {
	for (j = 0; j < mb; j++) {
	  C[i][j] = 0.0;
	  for (k = 0; k < na; k++) {
	    C[i][j] += A[i][k] * B[j][k];
	  }
	} 
      }
    }
  } 
  else {
    error_message("dmatmlt3","dimension inconsistent");
    printf("  A[%d][%d]  :  B[%d][%d] \n",ma,na,mb,nb);
    C = NULL;
  }
  return(C);
}


/**
 * \fn double *dmatdvct( double **A,  int m, int n, double *V, int mv)
 *
 * multiply matrix and vector (A*V)
 * \param A matrix
 * \param m number of rows of the matrix
 * \param n number of columns of the matrix
 * \param mv number of columns of the vector
 */
/********************************************************/
/*       <= A * V    (double matrix and double vector) 	*/
/*======================================================*/
/*	Usage: VCT = dmatdvct(A,m,n,V,mv);		*/
/********************************************************/
double *dmatdvct(
		 double **A,
		 int m, int n,
		 double *V,
		 int mv)
{
  double *VCT;
  int	 i,j;

  if (n == mv) {
    VCT = allocdvct(m);
    if (VCT != NULL) {
      for (i = 0; i < m; i++) {
	VCT[i] = 0.0;
	for (j = 0; j < n; j++) {
	  VCT[i] += A[i][j] * V[j];
	}
      }  
    }
  }
  else {
    error_message("rmatrvct","dimension inconsistent");
    printf(" A[%d][%d]  :  V[%d] \n",m,n,mv);
    VCT = NULL; /* debugged V -> VCT 1994 3 15 */
  }
  return(VCT); /* debugged V -> VCT 1994 3 15 */
}

