/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/

#ifndef lint
static char mutation_c_rcsid[]="$Id: mutation.cpp,v 1.2 2006/09/02 22:33:04 yanase Exp $";
#endif

/*
 *
 * $Log: mutation.cpp,v $
 * Revision 1.2  2006/09/02 22:33:04  yanase
 * add rename_tree_hold_type() in treegen.cpp/proto.h
 * to maintain the function node type after crossover and mutation.
 * replaced rename_tree by rename_tree_hold_type().
 * see also crossover.cpp and mutation.cpp.
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.4  2006/01/30 09:49:24  caranha
 * Added some better commenting
 *
 * Revision 1.3  2006/01/25 11:18:11  caranha
 * Added class for parameter manipulation. Moved some globals around.
 *
 * Revision 1.2  2006/01/24 07:45:51  caranha
 * Nothing really important, just moving globals around
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.5  1993/04/22  07:39:12  gpc-avc
 * Removed old log messages
 *
 * Revision 2.4  1993/04/14  04:54:06  gpc-avc
 * Finished mods for checkpointing
 * added format to write tree
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "gpc.h"
#include "proto.h"

#include "treeio.h"
#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

/* #define TRACEm */
tree *mutate(
  pop_struct	*pop,
  tree 		*t,
  int           *jcross
  )
{
  int	mpt;
  tree	*st, **stptr, *newst, *parent;
    
  nmut = nmut + 1;

  mpt = random_int(count_crossover_pts(t));

  *jcross = mpt;

#ifdef TRACEm
  printf("mpt = %d\n",mpt);
#endif

  st = get_subtree(t, mpt);

#ifdef TRACEm
  printf("subtree to be replaced:\n");
  write_tree(pop, st, pop[p].format, stdout);
#endif

  stptr = pointer_to_subtree(&t, st);
  
  /* note: unlike Koza/Rice we call this with root_p == 0, allowing a
     leaf to be substituted for the subtree we're snipping out */
  newst = create_random_tree(pop, t->pop, pop[t->pop].max_mutant_depth, 0, 0, NULL);

#ifdef TRACEm
  printf("new subtree:\n");
  write_tree(pop, newst, pop[p].format, stdout);
#endif

  parent = get_parent(t,st);

  free_tree(st);
  *stptr = newst;
  /* we return the value of t, since it may have changed */


  newst->parent = parent;
  //  rename_tree(newst);
  rename_tree_hold_type(newst); // to hold the type of function node, modified by yanase Sept 3, 2006

  return t;
}
#undef TRACEm
