#include <stdio.h>

#include <string.h>

#include "gpc.h"
#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"


problem::problem() // default problemwide values
{
  runtime = 0;

  numpops = 1;
  numgens = 200;

  data_max = 0;
  data_train_max = 0;
  data_gap = 119; // this is the default value for Nabatame's data
  test_and_train_same = 0;
  time_lag = 0;

  overlimit = 100;

  demes = 0;
  demerows = 0;
  demecols = 0;

  weight=1.0;
  weight_error=1.0;
  reg_param=0; // this must be bigger than 0 for regularization to have any effect

  eval_by_mdl=1;
};

void problem::inputTestData(double *newdata, int array_size)
{
  // If there is no train data, data_train_max should be 0
  for (int i=0; i<array_size; i++)
  {
    data[i+data_train_max] = newdata[i];
  }
  data_max = array_size+data_train_max;
};

void problem::inputTrainData(double *newdata, int array_size)
{
  double tmp[F_MAX]; // newdata MIGHT be just the needed size

  for (int i=0; i < array_size; i++)
  {
    tmp[i] = newdata[i];
    
  }
  if (data_train_max != data_max) // we have test data
  {
    for (int i=0; i < data_max - data_train_max; i++)
    {
      tmp[i+array_size] = data[i+data_train_max];
    }
  }
  memcpy(data,tmp,sizeof(data));
  data_max = data_max - data_train_max + array_size;
  data_train_max = array_size;

  
  getTrainVar();

};

void problem::inputTestDataFromFile(char *filename)
{
  double tmp[F_MAX];
  FILE *fi=fopen(filename,"r");
  int i=0;

  while (fscanf(fi,"%lf",&tmp[i])==1) // This stops on EOF or failure
  {
    i++;
  }

  inputTestData(tmp,i);
  fclose(fi);
};

void problem::inputTrainDataFromFile(char *filename)
{
  double tmp[F_MAX];
  FILE *fi=fopen(filename,"r");
  int i=0;

  while (fscanf(fi,"%lf",&tmp[i])==1) // This stops on EOF or failure
  {
    i++;
  }

  inputTrainData(tmp,i);
  fclose(fi);
};

void problem::normalize()
{

  double max=data[0];
  double min=data[0];

  memcpy(data_orig, data, sizeof(data));

  if (data_max == 0)
  {
    return;
  }

  for(int i=0; i< data_max ; i++)
  {
    if(data[i] > max)
    {
      max=data[i];
    }
    if(data[i] < min){
      min=data[i];
    }
  }
  for(int i=0; i< data_max ; i++)
  {
	data[i] = (data[i]-min)/(max-min);
  }
  
  getTrainVar();

}

// Calculates the training data variance
double problem::getTrainVar()
{
  double avg,dev;
  avg = 0;
  dev = 0;

  for (int i = 0; i < data_train_max; i++)
    {
      avg += data[i];
    }
  avg /= data_train_max;

  for (int i = 0; i < data_train_max; i++)
    {
      dev += (avg - data[i])*(avg - data[i]);
    }

  data_train_variance = (dev/data_train_max);

}

double problem::getTerminal(int current, int termindex)
{
  if ((current - termindex < 0)||
      (current - termindex > data_max))
  {
    return 0;
  }
  if (data[current - termindex] > P.overlimit) return P.overlimit;
  if (data[current - termindex] < -1*P.overlimit) return -1*P.overlimit;
  return data[current - termindex];
}
