/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/

#ifndef _PROTO_H_
#define _PROTO_H_




#ifndef lint
//static char proto_h_rcsid[]="$Id: proto.h,v 1.5 2006/09/04 09:42:40 caranha Exp $";
#endif

/*
 *
 * $Log: proto.h,v $
 * Revision 1.5  2006/09/04 09:42:40  caranha
 * Removed stupid mutate duplicates. Mutate offspring which is identical to parent
 *
 * Revision 1.4  2006/09/03 09:33:28  caranha
 * Addressing Identical Trees Problem - Not fully fixed yet
 *
 * Revision 1.3  2006/09/03 07:19:54  caranha
 * - Fixed clear_globals function so that it works. must be called after
 * running the program to prevent memory leaks for multiple runs/multiple
 * populations (globals.cpp/proto.h/main.cpp)
 *
 * - Removed some non-needed warning messages, replaced others to write to stderr
 * (gmdh.cpp)
 *
 * Revision 1.2  2006/09/02 22:33:04  yanase
 * add rename_tree_hold_type() in treegen.cpp/proto.h
 * to maintain the function node type after crossover and mutation.
 * replaced rename_tree by rename_tree_hold_type().
 * see also crossover.cpp and mutation.cpp.
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.2  2006/01/24 13:30:59  caranha
 * Modified "generations" to enable generation-per-generation running. Now you can stop one generation in order to pick it apart.
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.16  1993/04/29  20:39:23  gpc-avc
 * Added checks for NaN and infinity for results of eval
 *
 * Revision 2.15  1993/04/22  07:39:12  gpc-avc
 * Removed old log messages
 *
 * Revision 2.14  1993/04/15  09:14:39  gpc-avc
 * Added bsd_qsort()
 *
 *
 */

#if __STDC__ || defined(__cplusplus)
#define P_(s) s
#include <stddef.h>
#else
#define P_(s) ()
#endif

/* gpc.c */

void startup (
  int		argc,
  char		**argv,
  int		*numpops,
  int 		*numgens,
  int		*demes,
  int		*demerows,
  int		*demecols,
  int 		*start_gen,
  FILE		**ckpt_file,
  pop_struct 	**pop,
  pop_struct    ****grid
  );


void run_gp_system (int	numpops, int numgens, int start_gen,
  pop_struct 	*pop, int demes, pop_struct ***grid, int demerows,
  int demecols);




/* treeio.c */
int read_terminal_set_values P_((
  pop_struct *pop,
  int p,
  FILE *f
  ));
void load_terminal_set_values P_((
  pop_struct *pop,
  int p,
  //  float *v
  GENERIC *v
  ));
tree *read_tree P_((
  pop_struct *pop,
  int p,
  FILE *f
  ));

void write_tree P_((
  pop_struct *pop,
  tree *t,
  char *format,
  FILE *f
  ));

void write_tree_a P_((
  pop_struct *pop,
  tree *t,
  char *format,
  FILE *f
  ));

void pprint_tree P_((
  pop_struct *pop,
  tree *t,
  char *format,
  FILE *f,
  int level
  ));
int lookup_function_name P_((
  pop_struct *pop,
  int p,
  char *buf
  ));
int lookup_terminal_name P_((
  pop_struct *pop,
  int p,
  char *buf
  ));
void get_next_token P_((
  char *buf,
  FILE *f
  ));
void blanks P_((
  int n,
  FILE *f
  ));

void write_tailer(char *st);
void write_header(char *st);

/* treegen.c */
tree *create_random_tree P_((
  pop_struct *pop,
  int p,
  int max_depth,
  int root_p,
  int full_p,
  tree *parent
  ));

void rename_tree P_((
  tree *pos
  ));

// added by yanase Sept 3, 2006
void rename_tree_hold_type(tree *pos);


/* treeops.c */
tree *copy_tree P_((
  tree *t
  ));
void free_tree P_((
  tree *t
  ));
tree *create_tree_node P_((
  int pop,
  int nodetype,
  int id,
  tree *parent
  ));
void init_function P_((
  tree *t
  ));
void init_terminal P_((
  tree *t
  ));
tree *copy_function P_((
  tree *t,
  tree *parent
  ));
tree *copy_terminal P_((
  tree *t,
  tree *parent
  ));
void free_function P_((
  tree *t
  ));
void free_terminal P_((
  tree *t
  ));
int tree_cmp(tree *a, tree *b);


/* selection.c */
tree *find_tree P_((
  pop_struct 	*pop,
  int 		p,
  int 		demes,
  int 		nrows,
  int 		ncols,
  int		*worst,
  int		*best
  ));
tree *find_tree_using_demes P_((
  pop_struct 	*pop,
  int 		p,
  int 		nrows,
  int 		ncols,
  int		*worst,
  int		*best
  ));
float random_float_with_overselect P_((
  pop_struct *pop,
  int p
  ));
tree *find_tree_using_tournament_2 P_((
  pop_struct *pop,
  int p
  ));
tree *find_tree_using_tournament P_((
  pop_struct *pop,
  int 	p,
  int	*worst,
  int	*best
  ));
tree *find_tree_using_fitnessprop P_((
  pop_struct *pop,
  int p,
  float thresh,
  int		*best
  ));
int cannonical_select P_((
  pop_struct 	*pop,
  int 		p,
  int 		nrows,
  int 		ncols
  ));

/* crossover.c */
int count_crossover_pts P_((
  tree *t
  ));
void crossover_at_any_pt P_((
  pop_struct *pop,
  tree *p1,
  tree *p2,
  tree **o1,
  tree **o2,
  int           *jcross1,			 
  int           *jcross2

  ));
void crossover_at_func_pt P_((
  pop_struct *pop,
  tree *p1,
  tree *p2,
  tree **o1,
  tree **o2,
  int           *jcross1,			 
  int           *jcross2

  ));
void validate_crossover P_((
  pop_struct *pop,
  tree *p1,
  tree *p2,
  tree **o1,
  tree **o2,
  int           *flag1,
  int           *flag2

  ));
int depth_of_tree P_((
  tree *t
  ));
int count_crossover_pts P_((
  tree *t
  ));
int count_function_pts P_((
  tree *t
  ));
tree *get_subtree P_((
  tree *t,
  int n
  ));
tree *gs P_((
  tree *t,
  int *n
  ));
tree *get_function_subtree P_((
  tree *t,
  int n
  ));
tree *gfs P_((
  tree *t,
  int *n
  ));
tree **pointer_to_subtree P_((
  tree **pointer,
  tree *subt
  ));

/* populations.c */
void setup_deme_grid P_((
  int 		numpops,
  int		demerows,
  int		demecols,	       
  pop_struct 	*pop,
  pop_struct    ***grid
  ));
void allocate_populations P_((
  int numpops,
  pop_struct *pop
  ));
void initialize_populations P_((
  int numpops,
  pop_struct *pop
  ));
void breed_new_population P_((
  pop_struct *pop,
  int p,
  int demes,
  int nrows,
  int ncols,
  int steady_state			      
  ));
void free_population P_((
  pop_struct *pop,
  int p
  ));
void load_new_population P_((
  pop_struct *pop,
  int p
  ));

void sweep_pointer();
void clear_pointer();

/* generations.c */

void generations P_((
  int		numpops,
  int		numgens,
  int		start_gen,
  pop_struct	*pop,
  int		demes,
  pop_struct	***grid,
  int		demerows,
  int		demecols
  ));

void dump_population P_((
  pop_struct *pop,
  int p
  ));
void zero_fitness_of_populations P_((
  int	     numpops,
  pop_struct *pop,
  int	     p
  ));
void sort_population_by_fitness P_((
  pop_struct *pop,
  int p
  ));

void sort_population_by_mse P_((
  pop_struct *pop,
  int p
  ));

void sort_population_by_mdl P_((
  pop_struct *pop,
  int p
  ));

void add_parsimony_to_fitness P_((
  pop_struct *pop,
  int p
  ));
void normalize_fitness_of_population P_((
  pop_struct *pop,
  int p
  ));
void report_on_generation P_((
  int g,
  pop_struct *pop,
  int p
  ));

void write_predicted();
void free_tag(struct no_tag *tagp);
VOID gc_nodes(pop_struct *pop,int numpops);
VOID report_only_pop (
  int           pnum,
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st,
  int           flag
  );
VOID report (
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st,
  int           flag
  );



/* mutation.c */
tree *mutate P_((
  pop_struct 	*pop,
  tree 		*t,
  int           *jcross	    
  ));

/* eval.c */
GENERIC eval P_((
  tree *t
  ));

GENERIC evaluate_tree P_((
  tree *t
  ));

/* getparams.c */
void getparams P_((
  int		argc,
  char		**argv,
  int 		*numpops,
  int		*numgens,
  int		*demes,
  int		*demerows,
  int		*demecols,	       
  int 		*start_gen,
  FILE		**ckpt_file,
  pop_struct 	**pop,
  pop_struct    ****grid
  ));
void defaultparams P_((
  pop_struct *pop,
  int i
  ));
void readparams P_((
  pop_struct *pop,
  int i,
  FILE *f
  ));
void writeparams P_((
  pop_struct *pop,
  int i,
  FILE *f
  ));
void checkpoint P_((
  int           numpops,
  int           numgens,
  int           demes,
  int           demerows,
  int           demecols,
  pop_struct 	*pop,
  int		g,
  char          *st
  ));
void recover_params P_((
  FILE		*cf,
  int           *numpops,
  int           *numgens,
  int           *demes,
  int           *demerows,
  int           *demecols,
  pop_struct	**pop,
  pop_struct    ****grid,
  int		*g
  ));


/* fitness.c */
void evaluate_fitness_of_populations P_((
  int numpops,
  int numgens,
  pop_struct *pop,
  int p
  ));

double evaluate_fitness_of_individual P_((
     pop_struct	*pop,
     int	p,
     tree	*t,
     int	i
     ));


int terminate_early P_((
  int numpops,
  int numgens,
  pop_struct *pop
  ));

/* setup.c */
void make_function_table P_((
  int numpops,
  pop_struct *pop
  ));
void make_terminal_table P_((
  int numpops,
  pop_struct *pop
  ));

#include "random.h"



/* gmdh.c */

GENERIC mean_sqr_error P_((
  tree *t
  ));

int k_hat_new P_((tree *tr));

GENERIC aref_ind P_((int i, int ind));

GENERIC power P_((double x, int n));

GENERIC mdl P_((tree *tr, double sn2, int n));

GENERIC *b_calc P_((int n, int p));

tree *get_parent P_((tree *root, tree *me));

void fast_fit_gmdh P_((tree *t));

void test_data_plot P_((tree *t, FILE *cf));

/* Reads training and testing data set,
 * Must be changed later to comply with GUI
 */
void f_initialize_time ();

#endif

