/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi

 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved.

 - This notice may not be removed or altered.

 - You may not try to make money by distributing the package or by using the
   process that the code creates.

 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.

 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.

 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.

 - The authors are not responsible for the consequences of use of this
   software, no matter how awful, even if they arise from flaws in it.

If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/


/*
 *
 * $Log: setup.cpp,v $
 * Revision 1.4  2006/09/02 22:33:04  yanase
 * add rename_tree_hold_type() in treegen.cpp/proto.h
 * to maintain the function node type after crossover and mutation.
 * replaced rename_tree by rename_tree_hold_type().
 * see also crossover.cpp and mutation.cpp.
 *
 * Revision 1.3  2006/09/01 02:55:32  yanase
 * comment out original stroganoff function table
 *
 * Revision 1.2  2006/08/31 12:26:36  yanase
 * add functions of Adaptive learning of polynomial networks pp31.
 * I mainly modified setup.cpp.
 *
 * Revision 1.1.1.1  2006/08/31 08:56:26  yanase
 * This is the implementation of Inductive GP written in Adaptive Learning of Polynomial Networks.
 *
 * Revision 1.1.1.1  2006/08/21 09:52:44  caranha
 * Uploaded strogonoff Unix 02/2006 version to the server
 *
 * Revision 1.5  2006/01/31 05:49:00  caranha
 * *** empty log message ***
 *
 * Revision 1.4  2006/01/30 06:42:16  caranha
 * Quick dirty fix for the overflow problem - Artificial limiting parameters
 *
 * Revision 1.3  2006/01/27 11:31:38  caranha
 * fixed stupid bug at set local params
 *
 * Revision 1.2  2006/01/25 11:18:11  caranha
 * Added class for parameter manipulation. Moved some globals around.
 *
 * Revision 1.1  2006/01/24 03:54:24  caranha
 * Added initial files - stroglib is the library, debug-app is a debug application for the library - should not be needed
 *
 * Revision 2.12  1993/05/14  21:41:39  gpc-avc
 * Changed pdiv to return 1 on div by zero (per Koza)
 *
 * Revision 2.11  1993/04/30  05:01:48  gpc-avc
 * Restructured directories and Makefile
 *
 * Revision 2.9  1993/04/15  09:15:30  gpc-avc
 * deleted code for that was added for regression testing
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gpc.h"
#include "prob.h"
#include "gmdh.h"
#include "proto.h"

#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"

VOID make_function_table(int numpops, pop_struct *pop)
{

  /*
    Assigns function pointers, print names, arity, macro-flag, etc, to
    function table entries for all populations.  In this case all
    populations share the same function set.
  */

/* next is dummy */

  int	p;

  for (p=0; p<numpops; p++) {
    pop[p].function_table_size = 16;
    pop[p].function_table =
      (function_table_entry *) malloc(pop[p].function_table_size *
				      sizeof(function_table_entry));

    sprintf(pop[p].function_table[0].printname,"+");

  }

  no_node_functions = 16;

/* always begins at 1, not 0  for -1,-2... */
/*
  node_functions[1].arity1 = 6;
  node_functions[1].arity2 = 3;
  sprintf(node_functions[1].printname,"Quadratic");
  node_functions[1].code1 = quad1;
  node_functions[1].code2 = quad2;
*/
  // add by yanase 31 Aug, 2006
  node_functions[1].arity1 = 4;
  node_functions[1].arity2 = 4;
  sprintf(node_functions[1].printname,"P1");
  node_functions[1].code1 = polynomial1;
  node_functions[1].code2 = polynomial1;

  node_functions[2].arity1 = 3;
  node_functions[2].arity2 = 3;
  sprintf(node_functions[2].printname,"P2");
  node_functions[2].code1 = polynomial2;
  node_functions[2].code2 = polynomial2;

  node_functions[3].arity1 = 3;
  node_functions[3].arity2 = 3;
  sprintf(node_functions[3].printname,"P3");
  node_functions[3].code1 = polynomial3;
  node_functions[3].code2 = polynomial3;

  node_functions[4].arity1 = 4;
  node_functions[4].arity2 = 4;
  sprintf(node_functions[4].printname,"P4");
  node_functions[4].code1 = polynomial4;
  node_functions[4].code2 = polynomial4;

  node_functions[5].arity1 = 3;
  node_functions[5].arity2 = 3;
  sprintf(node_functions[5].printname,"P5");
  node_functions[5].code1 = polynomial5;
  node_functions[5].code2 = polynomial5;

  node_functions[6].arity1 = 4;
  node_functions[6].arity2 = 4;
  sprintf(node_functions[6].printname,"P6");
  node_functions[6].code1 = polynomial6;
  node_functions[6].code2 = polynomial6;

  node_functions[7].arity1 = 4;
  node_functions[7].arity2 = 4;
  sprintf(node_functions[7].printname,"P7");
  node_functions[7].code1 = polynomial7;
  node_functions[7].code2 = polynomial7;

  node_functions[8].arity1 = 3;
  node_functions[8].arity2 = 3;
  sprintf(node_functions[8].printname,"P8");
  node_functions[8].code1 = polynomial8;
  node_functions[8].code2 = polynomial8;

  node_functions[9].arity1 = 6;
  node_functions[9].arity2 = 6;
  sprintf(node_functions[9].printname,"P9");
  node_functions[9].code1 = polynomial9;
  node_functions[9].code2 = polynomial9;

  node_functions[10].arity1 = 5;
  node_functions[10].arity2 = 5;
  sprintf(node_functions[10].printname,"P10");
  node_functions[10].code1 = polynomial10;
  node_functions[10].code2 = polynomial10;

  node_functions[11].arity1 = 5;
  node_functions[11].arity2 = 5;
  sprintf(node_functions[11].printname,"P11");
  node_functions[11].code1 = polynomial11;
  node_functions[11].code2 = polynomial11;

  node_functions[12].arity1 = 4;
  node_functions[12].arity2 = 4;
  sprintf(node_functions[12].printname,"P12");
  node_functions[12].code1 = polynomial12;
  node_functions[12].code2 = polynomial12;

  node_functions[13].arity1 = 4;
  node_functions[13].arity2 = 4;
  sprintf(node_functions[13].printname,"P13");
  node_functions[13].code1 = polynomial13;
  node_functions[13].code2 = polynomial13;

  node_functions[14].arity1 = 5;
  node_functions[14].arity2 = 5;
  sprintf(node_functions[14].printname,"P14");
  node_functions[14].code1 = polynomial14;
  node_functions[14].code2 = polynomial14;

  node_functions[15].arity1 = 2;
  node_functions[15].arity2 = 2;
  sprintf(node_functions[15].printname,"P15");
  node_functions[15].code1 = polynomial15;
  node_functions[15].code2 = polynomial15;

  node_functions[16].arity1 = 3;
  node_functions[16].arity2 = 3;
  sprintf(node_functions[16].printname,"P16");
  node_functions[16].code1 = polynomial16;
  node_functions[16].code2 = polynomial16;
}

VOID make_terminal_table(int numpops, pop_struct *pop)
{
  /*
     IMPORTANT NOTE: terminal_table_size should be equal to the number of
     terminals, but there should be (terminal_table_size+1) entries to allow
     for reading/printing actual constants
   */


  int p,i;
  /*
    NOTE: all populations are the same for this case
  */


  for (p=0; p<numpops; p++) {

    pop[p].terminal_table =
      (terminal_table_entry *) malloc((pop[p].terminal_table_size+1) *
				      sizeof(terminal_table_entry));
  for (i=0; i<pop[p].terminal_table_size; i++)
  {
    pop[p].terminal_table[i].val = 0;
    sprintf(pop[p].terminal_table[i].printname, "X%d",i+1);
    pop[p].terminal_table[i].constant_generator = random_constant;
  }

//----------------------------------------------------------------------
/* not used  for ALLOW_CONST = 0 */
    pop[p].terminal_table[pop[p].terminal_table_size].val = 0;
    strcpy(pop[p].terminal_table[pop[p].terminal_table_size].printname,FORMAT);
    pop[p].terminal_table[pop[p].terminal_table_size].constant_generator =
      random_constant;

    pop[p].format = FORMAT;
    pop[p].ckpt_format = CKPT_FORMAT;

  }
}

/**
 * \fn GENERIC *quad1(GENERIC *args)
 *
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *quad1(GENERIC *args)
{
  double *ans;

  ans = (double *)allocvct(6,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[1];
  ans[4] = args[0]*args[0];
  ans[5] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < 6; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *quad2(GENERIC *args)
 *
 * \brief function of non-terminal node
 * \param *args x1
 * \return array to calculate coefficients
 */
GENERIC *quad2(GENERIC *args)
{
  double *ans;

  ans = (double *)allocvct(3,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[0];
  for (int i = 0; i < 3; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

GENERIC random_constant(void)
{
  return ((GENERIC)random_float(10.0) - (GENERIC)5.0);
}

/// added by yanase 31 Aug, 2006
/// based on Adaptive learning of Polynomial Networks pp.31 
/// to update this program as iGP

/**
 * \fn GENERIC *polynomial1(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial1(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial2(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial2(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 3;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial3(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial3(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 3;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial4(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial4(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[1];
  ans[3] = args[0]*args[0];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial5(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial5(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 3;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial6(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial6(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[0];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial7(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial7(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[0];
  ans[3] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial8(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial8(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 3;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0]*args[0];
  ans[2] = args[1]*args[0];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial9(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial9(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 6;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[1];
  ans[4] = args[0]*args[0];
  ans[5] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial10(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial10(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 5;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[1];
  ans[4] = args[0]*args[0];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial11(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial11(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 5;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[1];
  ans[3] = args[0]*args[0];
  ans[4] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial12(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial12(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0]*args[1];
  ans[2] = args[0]*args[0];
  ans[3] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial13(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial13(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 4;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[0]*args[1];
  ans[3] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial14(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial14(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 5;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0];
  ans[2] = args[1];
  ans[3] = args[0]*args[0];
  ans[4] = args[1]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial15(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial15(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 2;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0]*args[1];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}

/**
 * \fn GENERIC *polynomial16(GENERIC *args)
 *
 * @author Toshihiko YANASE
 * \brief function of non-terminal node
 * \param *args x1,x2
 * \return array to calculate coefficients
 */
GENERIC *polynomial16(GENERIC *args)
{
  double *ans;
  int num_of_coeff = 3;

  ans = (double *)allocvct(num_of_coeff,sizeof(double));
  ans[0] = 1.0;
  ans[1] = args[0]*args[1];
  ans[2] = args[0]*args[0];

  // Kludge for preventing overflow. Should do this better
  for (int i = 0; i < num_of_coeff; i++)
  {
    if (ans[i] > P.overlimit) ans[i] = P.overlimit;
    if (ans[i] < -1*P.overlimit) ans[i] = -1*P.overlimit;
  }
  return(ans);
}


/**
 * \fn char *test_poly_sub(int index)
 * \breif test function of polynomial?
 * convert number which is the result of polynomial? function to ascii.
 *
 * @author Toshihiko Yanase
 * \param index number of calculated by polynomial?
 */
char *test_poly_sub(int index){
  switch(index){
  case 1:
    return "1";
  case 2:
    return "x1";
  case 3:
    return "x2";
  case 4:
    return "x1^2";
  case 6:
    return "x1x2";
  case 9:
    return "x2^2";
  default:
    return "error";
  }
}

/**
 * \fn char *test_poly(int index)
 * \breif test function of polynomial?
 *
 * @author Toshihiko Yanase
 */
void test_poly(){
  double args[2];
  args[0]=2;
  args[1]=3;

  double *ans;
  int num_of_ans[] = {4,3,3,4,3,4,4,3,6,5,5,4,4,5,2,3};
  double * (*p_func[])(double *)={
    polynomial1,
    polynomial2,
    polynomial3,
    polynomial4,
    polynomial5,
    polynomial6,
    polynomial7,
    polynomial8,
    polynomial9,
    polynomial10,
    polynomial11,
    polynomial12,
    polynomial13,
    polynomial14,
    polynomial15,
    polynomial16
  };

  int i,j;
  for(i=0;i<16;i++){
    ans = p_func[i](args);
    printf("%d: ",i+1);
    for(j=0;j<num_of_ans[i];j++){
      //      printf("%f,", ans[j]);
      printf("%s ", test_poly_sub((int) ans[j]));
    }
    deallocvct(ans);
    printf("\n");
  }

}
