/*
SGPC: Simple Genetic Programming in C
(c) 1993 by Walter Alden Tackett and Aviram Carmi
 
 This code and documentation is copyrighted and is not in the public domain.
 All rights reserved. 
 
 - This notice may not be removed or altered.
 
 - You may not try to make money by distributing the package or by using the
   process that the code creates.
 
 - You may not distribute modified versions without clearly documenting your
   changes and notifying the principal author.
 
 - The origin of this software must not be misrepresented, either by
   explicit claim or by omission.  Since few users ever read sources,
   credits must appear in the documentation.
 
 - Altered versions must be plainly marked as such, and must not be
   misrepresented as being the original software.  Since few users ever read
   sources, credits must appear in the documentation.
 
 - The authors are not responsible for the consequences of use of this 
   software, no matter how awful, even if they arise from flaws in it.
 
If you make changes to the code, or have suggestions for changes,
let us know!  (gpc@ipld01.hac.com)
*/



#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "gpc.h"
#include "proto.h"

#include "magic.h"
#include "problem.h"
#include "getparams.h"
#include "globals.h"


tree *create_tree_node(
  int pop,
  int nodetype,
  int id,
  tree *parent
  )
{
  tree	*t;
  
  if ((t = (tree *) malloc(sizeof(tree))) == NULL) {
    perror(" at point 1 in create_tree_node() <treeops.c>");
    exit(errno);
  }
  
  t->nodetype = nodetype;
  t->id = id;
  t->pop = pop;
  t->parent = parent;

  switch (nodetype) {
  case FUNCTION:
    init_function(t);
    break;
  case TERMINAL:
/*     init_terminal(t); */
    /* nothing to do */
    break;
  default:
    fprintf(stderr,
	    "nodetype %d must be %d or %d in create_tree_node() <treeops.c>\n",
	    nodetype, FUNCTION, TERMINAL);
  }
  
  return t;
}

VOID init_function(
  tree *t
  )
{
  pop_struct *pop = POP;

 
  if ((t->id < 0) || (t->id >= pop[t->pop].function_table_size)) {
    fprintf(stderr,
	    "id %d must be >= 0 and < %d in init_function() <treeops.c>\n",
	    t->id, pop[t->pop].function_table_size);
  }

  
  
  if ((t->type.func = (function *) malloc(sizeof(function))) == NULL) {
    perror(" at point 1 in init_function() <treeops.c>");
    exit(errno);
  }
  
  if ((t->type.func->arg =
       (tree **) malloc(function_arity(t) * sizeof(tree *))) == NULL) {
    perror(" at point 2 in init_function() <treeops.c>");
    exit(errno);
  }

}


tree *copy_tree_new(
	tree *t,
	tree *parent
	)
{
  switch (t->nodetype) {
  case FUNCTION:
    return copy_function(t,parent);
  case TERMINAL:
    return copy_terminal(t,parent);
  default:
    fprintf(stderr,
	    "nodetype %d must be %d or %d in copy_tree() <treeops.c>\n",
	    t->nodetype, FUNCTION, TERMINAL);
    return (tree *) NULL;
  }
}


tree *copy_tree(
	tree *t
	)
{
  return (copy_tree_new(t,NULL));
}



tree *copy_function(
  tree *t,
  tree *parent
  )
{
  tree *nt;
  int  i;

  nt = create_tree_node(t->pop,t->nodetype, t->id, parent);
  for (i=0; i < function_arity(t); i++ ) {
    nt->type.func->arg[i] = copy_tree_new(t->type.func->arg[i], nt);
  }
  nt -> type.func -> node_pointer = t-> type.func -> node_pointer;

  return nt;
}


tree *copy_terminal(
  tree *t,
  tree *parent
  )
{
  tree *nt;

  nt = create_tree_node(t->pop,t->nodetype, t->id, parent);
/*   if (terminal_is_constant(t)) {
    *(nt->type.term->valptr) = (*(t->type.term->valptr));
  } */
  return nt;
}


VOID free_tree(
  tree *t
  )
{
  switch (t->nodetype) {
  case FUNCTION:
    free_function(t);
    break;
  case TERMINAL:
    free_terminal(t);
    break;
  default:
    fprintf(stderr,
	    "nodetype %d must be %d or %d in free_tree() <treeops.c>\n",
	    t->nodetype, FUNCTION, TERMINAL);
  }
}

VOID free_function(
  tree *t
  )
{
  int  i;

  for (i=0; i < function_arity(t); i++ ) {
    free_tree(t->type.func->arg[i]);
  }
  free(t->type.func->arg);
/*   free(t->type.func->argvals); */
  free(t->type.func);
  free(t);
}

VOID free_terminal(
  tree *t
  )
{
/*  if (terminal_is_constant(t)) {
    free(t->type.term->valptr);
  } */
/*  free(t->type.term); */
  free(t);
}


/* Compares tree a and tree b, returning 0 if same, 1 if different */
int tree_cmp(tree *a, tree *b)
{
  int diff,i;
  //fprintf(stderr, "nodetype: %d %d\n", a->nodetype, b->nodetype);
  if (a->nodetype != b->nodetype) 
    return (1);
  if (a->nodetype == FUNCTION)
    {
      if (a->type.func->node_pointer->type !=
	  b->type.func->node_pointer->type)
	return(1);
      
      diff = 0;
      for (i=0; i<function_arity(t); i++)
	{
	  diff = diff || tree_cmp(a->type.func->arg[i], 
				  b->type.func->arg[i]);
	  if (diff == 1)
	    return (diff);
	}
      return (diff);
    }
  else if (a->nodetype == TERMINAL)
    {
      return(strcmp(terminal_printname(a),terminal_printname(b)));
    }
  
  fprintf(stderr,"Error in tree_cmp\n");
  return(0);

}
